import json
from abc import ABC, abstractmethod

from sirius_sdk.base import JsonSerializable


class CacheOptions(JsonSerializable):

    def __init__(self, no_cache: bool=False, no_update: bool=False, no_store: bool=False, min_fresh: int=-1):
        """
        :param no_cache: (bool, optional, false by default) Skip usage of cache,
        :param no_update: (bool, optional, false by default) Use only cached data, do not try to update.
        :param no_store: (bool, optional, false by default) Skip storing fresh data if updated,
        :param min_fresh: (int, optional, -1 by default) Return cached data if not older than this many seconds. -1 means do not check age.
        """
        self.no_cache = no_cache
        self.no_update = no_update
        self.no_store = no_store
        self.min_fresh = min_fresh

    def to_json(self):
        return {
           'noCache': self.no_cache,
           'noUpdate': self.no_update,
           'noStore': self.no_store,
           'minFresh': self.min_fresh
        }

    def serialize(self):
        return json.dumps(self.to_json())

    def deserialize(self, buffer: str):
        data = json.loads(buffer)
        self.no_cache = data.get('noCache', False)
        self.no_update = data.get('noUpdate', False)
        self.no_store = data.get('noStore', False)
        self.min_fresh = data.get('minFresh', -1)


class PurgeOptions(JsonSerializable):

    def __init__(self, max_age: int=-1):
        self.max_age = max_age

    def to_json(self):
        return {
           'maxAge': self.max_age
        }

    def serialize(self):
        return json.dumps(self.to_json())

    def deserialize(self, buffer: str):
        data = json.loads(buffer)
        self.max_age = data.get('maxAge', -1)


class AbstractCache(ABC):

    @abstractmethod
    async def get_schema(self, pool_name: str, submitter_did: str, id_: str, options: CacheOptions) -> dict:
        """
        Gets schema json data for specified schema id.
        If data is present inside of cache, cached data is returned.
        Otherwise data is fetched from the ledger and stored inside of cache for future use.

        EXPERIMENTAL

        :param pool_name: Ledger.
        :param submitter_did: DID of the submitter stored in secured Wallet.
        :param id_: identifier of schema.
        :param options:
        {
            noCache: (bool, optional, false by default) Skip usage of cache,
            noUpdate: (bool, optional, false by default) Use only cached data, do not try to update.
            noStore: (bool, optional, false by default) Skip storing fresh data if updated,
            minFresh: (int, optional, -1 by default) Return cached data if not older than this many seconds. -1 means do not check age.
        }
        :return: Schema json.
        {
            id: identifier of schema
            attrNames: array of attribute name strings
            name: Schema's name string
            version: Schema's version string
            ver: Version of the Schema json
        }
        """
        raise NotImplemented

    @abstractmethod
    async def get_cred_def(self, pool_name: str, submitter_did: str, id_: str, options: CacheOptions) -> dict:
        """
        Gets credential definition json data for specified credential definition id.
        If data is present inside of cache, cached data is returned.
        Otherwise data is fetched from the ledger and stored inside of cache for future use.

        EXPERIMENTAL

        :param pool_name: Ledger.
        :param submitter_did: DID of the submitter stored in secured Wallet.
        :param id_: identifier of credential definition.
        :param options:
        {
            noCache: (bool, optional, false by default) Skip usage of cache,
            noUpdate: (bool, optional, false by default) Use only cached data, do not try to update.
            noStore: (bool, optional, false by default) Skip storing fresh data if updated,
            minFresh: (int, optional, -1 by default) Return cached data if not older than this many seconds. -1 means do not check age.
        }
        :return: Credential Definition json.
        {
            id: string - identifier of credential definition
            schemaId: string - identifier of stored in ledger schema
            type: string - type of the credential definition. CL is the only supported type now.
            tag: string - allows to distinct between credential definitions for the same issuer and schema
            value: Dictionary with Credential Definition's data: {
                primary: primary credential public key,
                Optional<revocation>: revocation credential public key
            },
            ver: Version of the Credential Definition json
        }
        """
        raise NotImplemented

    @abstractmethod
    async def purge_schema_cache(self, options: PurgeOptions) -> None:
        """
        Purge schema cache.

        EXPERIMENTAL

        :param options:
        {
            maxAge: (int, optional, -1 by default) Purge cached data if older than this many seconds. -1 means purge all.
        }
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def purge_cred_def_cache(self, options: PurgeOptions) -> None:
        """
        Purge credential definition cache.

        EXPERIMENTAL

        :param options:
        {
            maxAge: (int, optional, -1 by default) Purge cached data if older than this many seconds. -1 means purge all.
        }
        :return: None
        """
        raise NotImplemented



from sirius_sdk.agent.aries_rfc.feature_0160_connection_protocol.messages import Invitation, ConnRequest
from sirius_sdk.agent.aries_rfc.feature_0160_connection_protocol.state_machines import Inviter, Invitee


__all__ = ['Invitation', 'ConnRequest', 'Inviter', 'Invitee']



import base64
from typing import Any

from sirius_sdk.agent.wallet import CacheOptions, PurgeOptions, RetrieveRecordOptions, NYMRole, PoolAction, KeyDerivationMethod
from sirius_sdk.rpc.futures import Future
from sirius_sdk.messaging import Message, Type
from sirius_sdk.errors.exceptions import SiriusInvalidType, SiriusInvalidPayloadStructure


MSG_TYPE_FUTURE = 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/future'


CLS_MAP = {
    'application/cache-options': CacheOptions,
    'application/purge-options': PurgeOptions,
    'application/retrieve-record-options': RetrieveRecordOptions,
    'application/nym-role': NYMRole,
    'application/pool-action': PoolAction,
    'application/key-derivation-method': KeyDerivationMethod,
}
CLS_MAP_REVERT = {v: k for k, v in CLS_MAP.items()}


def serialize_variable(var: Any) -> (str, Any):
    """Serialize input variable to JSON-compatible string

    :param var: input variable
    :return: tuple (type, variable serialized dump)
    """
    if isinstance(var, CacheOptions):
        return CLS_MAP_REVERT[CacheOptions], var.serialize()
    elif isinstance(var, PurgeOptions):
        return CLS_MAP_REVERT[PurgeOptions], var.serialize()
    elif isinstance(var, RetrieveRecordOptions):
        return CLS_MAP_REVERT[RetrieveRecordOptions], var.serialize()
    elif isinstance(var, NYMRole):
        return CLS_MAP_REVERT[NYMRole], var.serialize()
    elif isinstance(var, PoolAction):
        return CLS_MAP_REVERT[PoolAction], var.serialize()
    elif isinstance(var, KeyDerivationMethod):
        return CLS_MAP_REVERT[KeyDerivationMethod], var.serialize()
    elif isinstance(var, bytes):
        return 'application/base64', base64.b64encode(var).decode('ascii')
    else:
        return None, var


def deserialize_variable(payload: Any, typ: str=None) -> Any:
    """Deserialize variable from string according to type

    :param payload: input variable
    :param typ: variable type
    :return: deserialized variable
    """
    if typ is None:
        return payload
    elif typ == 'application/base64':
        return base64.b64decode(payload.encode('ascii'))
    elif typ in CLS_MAP.keys():
        cls = CLS_MAP[typ]
        if issubclass(cls, NYMRole):
            inst = NYMRole.deserialize(payload)
        elif issubclass(cls, PoolAction):
            inst = PoolAction.deserialize(payload)
        elif issubclass(cls, KeyDerivationMethod):
            inst = KeyDerivationMethod.deserialize(payload)
        else:
            inst = cls()
            inst.deserialize(payload)
        return inst
    else:
        raise RuntimeError('Unexpected typ: "%s"' % typ)


def incapsulate_param(param):
    typ, payload = serialize_variable(param)
    return {
        'mime_type': typ,
        'payload': payload
    }


def deincapsulate_param(packet: dict):
    return deserialize_variable(packet['payload'], packet['mime_type'])


def build_request(msg_type: str, future: Future, params: dict) -> Message:
    """

    :param msg_type: Aries RFCs attribute
        https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0020-message-types
    :param future: Future to check response routine is completed
    :param params: RPC call params
    :return: RPC service packet
    """
    typ = Type.from_str(msg_type)
    if typ.protocol not in ['sirius_rpc', 'admin', 'microledgers', 'microledgers-batched']:
        raise SiriusInvalidType('Expect sirius_rpc protocol')
    return Message({
        '@type': msg_type,
        '@promise': future.promise,
        'params': {k: incapsulate_param(v) for k, v in params.items()}
    })


def build_response(packet: Message):
    if packet.get('@type') == MSG_TYPE_FUTURE:
        if packet.get('~thread', None) is not None:
            parsed = {
                'exception': None,
                'value': None
            }
            exception = packet['exception']
            if exception:
                parsed['exception'] = exception
            else:
                value = packet['value']
                if packet['is_tuple']:
                    parsed['value'] = tuple(value)
                elif packet['is_bytes']:
                    parsed['value'] = base64.b64decode(value.encode('ascii'))
                else:
                    parsed['value'] = value
            return parsed
        else:
            raise SiriusInvalidPayloadStructure('Expect ~thread decorator')
    else:
        raise SiriusInvalidType('Expect message type "%s"' % MSG_TYPE_FUTURE)



from typing import List, Optional

from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds, AnonCredSchema
from sirius_sdk.agent.connections import AgentRPC


class AnonCredsProxy(AbstractAnonCreds):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def issuer_create_schema(
            self, issuer_did: str, name: str, version: str, attrs: List[str]
    ) -> (str, AnonCredSchema):
        schema_id, body = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_create_schema',
            params=dict(issuer_did=issuer_did, name=name, version=version, attrs=attrs)
        )
        return schema_id, AnonCredSchema(**body)

    async def issuer_create_and_store_credential_def(
            self, issuer_did: str, schema: dict, tag: str, signature_type: str = None, config: dict = None
    ) -> (str, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_create_and_store_credential_def',
            params=dict(issuer_did=issuer_did, schema=schema, tag=tag, signature_type=signature_type, config=config)
        )

    async def issuer_rotate_credential_def_start(self, cred_def_id: str, config: dict = None) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_rotate_credential_def_start',
            params=dict(cred_def_id=cred_def_id, config=config)
        )

    async def issuer_rotate_credential_def_apply(self, cred_def_id: str):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_rotate_credential_def_apply',
            params=dict(cred_def_id=cred_def_id)
        )

    async def issuer_create_and_store_revoc_reg(
            self, issuer_did: str, revoc_def_type: Optional[str], tag: str,
            cred_def_id: str, config: dict, tails_writer_handle: int
    ) -> (str, dict, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_create_and_store_revoc_reg',
            params=dict(
                issuer_did=issuer_did, revoc_def_type=revoc_def_type, tag=tag,
                cred_def_id=cred_def_id, config=config, tails_writer_handle=tails_writer_handle
            )
        )

    async def issuer_create_credential_offer(self, cred_def_id: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_create_credential_offer',
            params=dict(cred_def_id=cred_def_id)
        )

    async def issuer_create_credential(
            self, cred_offer: dict, cred_req: dict, cred_values: dict,
            rev_reg_id: str = None, blob_storage_reader_handle: int = None
    ) -> (dict, Optional[str], Optional[dict]):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_create_credential',
            params=dict(
                cred_offer=cred_offer, cred_req=cred_req, cred_values=cred_values,
                rev_reg_id=rev_reg_id, blob_storage_reader_handle=blob_storage_reader_handle
            )
        )

    async def issuer_revoke_credential(
            self, blob_storage_reader_handle: int, rev_reg_id: str, cred_revoc_id: str
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_revoke_credential',
            params=dict(
                blob_storage_reader_handle=blob_storage_reader_handle,
                rev_reg_id=rev_reg_id, cred_revoc_id=cred_revoc_id
            )
        )

    async def issuer_merge_revocation_registry_deltas(self, rev_reg_delta: dict, other_rev_reg_delta: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/issuer_merge_revocation_registry_deltas',
            params=dict(rev_reg_delta=rev_reg_delta, other_rev_reg_delta=other_rev_reg_delta)
        )

    async def prover_create_master_secret(self, master_secret_name: str = None) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_create_master_secret',
            params=dict(master_secret_name=master_secret_name)
        )

    async def prover_create_credential_req(
            self, prover_did: str, cred_offer: dict, cred_def: dict, master_secret_id: str
    ) -> (dict, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_create_credential_req',
            params=dict(
                prover_did=prover_did, cred_offer=cred_offer, cred_def=cred_def, master_secret_id=master_secret_id
            )
        )

    async def prover_set_credential_attr_tag_policy(
            self, cred_def_id: str, tag_attrs: Optional[dict], retroactive: bool
    ) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_set_credential_attr_tag_policy',
            params=dict(cred_def_id=cred_def_id, tag_attrs=tag_attrs, retroactive=retroactive)
        )

    async def prover_get_credential_attr_tag_policy(self, cred_def_id: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_get_credential_attr_tag_policy',
            params=dict(cred_def_id=cred_def_id)
        )

    async def prover_store_credential(
            self, cred_id: Optional[str], cred_req_metadata: dict, cred: dict, cred_def: dict, rev_reg_def: dict = None
    ) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_store_credential',
            params=dict(
                cred_id=cred_id, cred_req_metadata=cred_req_metadata,
                cred=cred, cred_def=cred_def, rev_reg_def=rev_reg_def
            )
        )

    async def prover_get_credential(self, cred_id: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_get_credential',
            params=dict(cred_id=cred_id)
        )

    async def prover_delete_credential(self, cred_id: str) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_delete_credential',
            params=dict(cred_id=cred_id)
        )

    async def prover_get_credentials(self, filters: dict) -> List[dict]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_get_credentials',
            params=dict(filters=filters)
        )

    async def prover_search_credentials(self, query: dict) -> List[dict]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_search_credentials',
            params=dict(query=query)
        )

    async def prover_get_credentials_for_proof_req(self, proof_request: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_get_credentials_for_proof_req',
            params=dict(proof_request=proof_request)
        )

    async def prover_search_credentials_for_proof_req(
            self, proof_request: dict, extra_query: dict = None, limit_referents: int = 1
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_search_credentials_for_proof_req',
            params=dict(
                proof_request=proof_request, extra_query=extra_query, limit_referents=limit_referents
            )
        )

    async def prover_create_proof(
            self, proof_req: dict, requested_credentials: dict, master_secret_name: str,
            schemas: dict, credential_defs: dict, rev_states: dict
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prover_create_proof',
            params=dict(
                proof_req=proof_req, requested_credentials=requested_credentials,
                master_secret_name=master_secret_name, schemas=schemas,
                credential_defs=credential_defs, rev_states=rev_states
            )
        )

    async def verifier_verify_proof(
            self, proof_request: dict, proof: dict, schemas: dict,
            credential_defs: dict, rev_reg_defs: dict, rev_regs: dict
    ) -> bool:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/verifier_verify_proof',
            params=dict(
                proof_request=proof_request, proof=proof, schemas=schemas,
                credential_defs=credential_defs, rev_reg_defs=rev_reg_defs, rev_regs=rev_regs
            )
        )

    async def create_revocation_state(
            self, blob_storage_reader_handle: int, rev_reg_def: dict,
            rev_reg_delta: dict, timestamp: int, cred_rev_id: str
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/create_revocation_state',
            params=dict(
                blob_storage_reader_handle=blob_storage_reader_handle, rev_reg_def=rev_reg_def,
                rev_reg_delta=rev_reg_delta, timestamp=timestamp, cred_rev_id=cred_rev_id
            )
        )

    async def update_revocation_state(
            self, blob_storage_reader_handle: int, rev_state: dict, rev_reg_def: dict,
            rev_reg_delta: dict, timestamp: int, cred_rev_id: str
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/update_revocation_state',
            params=dict(
                blob_storage_reader_handle=blob_storage_reader_handle, rev_state=rev_state,
                rev_reg_def=rev_reg_def, rev_reg_delta=rev_reg_delta, timestamp=timestamp, cred_rev_id=cred_rev_id
            )
        )

    async def generate_nonce(self) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/generate_nonce'
        )

    async def to_unqualified(self, entity: str) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/to_unqualified',
            params=dict(entity=entity)
        )



import os
import uuid
import asyncio

import pytest

from sirius_sdk import Agent, Pairwise
from sirius_sdk.rpc import AddressedTunnel
from sirius_sdk.encryption import create_keypair, bytes_to_b58, P2PConnection
from .helpers import InMemoryChannel, ServerTestSuite, IndyAgent


SERVER_SUITE = None
INDY_AGENT = None


def pytest_configure():
    # Address of TestSuite
    pytest.test_suite_baseurl = os.getenv('TEST_SUITE_BASE_URL') or 'http://localhost'
    pytest.test_suite_overlay_address = 'http://10.0.0.90'
    # Back compatibility testing
    pytest.old_agent_address = os.getenv('INDY_AGENT_BASE_URL') or 'http://127.0.0.1:88'
    pytest.old_agent_overlay_address = 'http://10.0.0.52:8888'
    pytest.old_agent_root = {
        'username': 'root',
        'password': 'root'
    }


@pytest.fixture()
def p2p() -> dict:
    keys_agent = create_keypair(b'000000000000000000000000000AGENT')
    keys_sdk = create_keypair(b'00000000000000000000000000000SDK')
    agent = P2PConnection(
        my_keys=(
            bytes_to_b58(keys_agent[0]),
            bytes_to_b58(keys_agent[1])
        ),
        their_verkey=bytes_to_b58(keys_sdk[0])
    )
    smart_contract = P2PConnection(
        my_keys=(
            bytes_to_b58(keys_sdk[0]),
            bytes_to_b58(keys_sdk[1])
        ),
        their_verkey=bytes_to_b58(keys_agent[0])
    )
    downstream = InMemoryChannel()
    upstream = InMemoryChannel()
    return {
        'agent': {
            'p2p': agent,
            'tunnel': AddressedTunnel('memory://agent->sdk', upstream, downstream, agent)
        },
        'sdk': {
            'p2p': smart_contract,
            'tunnel': AddressedTunnel('memory://sdk->agent', downstream, upstream, smart_contract)
        }
    }


def get_suite_singleton() -> ServerTestSuite:
    global SERVER_SUITE
    if not isinstance(SERVER_SUITE, ServerTestSuite):
        suite = ServerTestSuite()
        asyncio.get_event_loop().run_until_complete(suite.ensure_is_alive())
        SERVER_SUITE = suite
    return SERVER_SUITE


def get_indy_agent_singleton() -> IndyAgent:
    global INDY_AGENT
    if not isinstance(INDY_AGENT, IndyAgent):
        agent = IndyAgent()
        asyncio.get_event_loop().run_until_complete(agent.ensure_is_alive())
        INDY_AGENT = agent
    return INDY_AGENT


def get_agent(name: str) -> Agent:
    params = get_suite_singleton().get_agent_params(name)
    agent = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=30,
        name=name
    )
    return agent


@pytest.fixture()
def test_suite() -> ServerTestSuite:
    return get_suite_singleton()


@pytest.fixture()
def prover_master_secret_name() -> str:
    return 'prover_master_secret_name'


@pytest.fixture()
def indy_agent() -> IndyAgent:
    return get_indy_agent_singleton()


@pytest.fixture()
def agent1() -> Agent:
    return get_agent('agent1')


@pytest.fixture()
def agent2() -> Agent:
    return get_agent('agent2')


@pytest.fixture()
def agent3() -> Agent:
    return get_agent('agent3')


@pytest.fixture()
def agent4() -> Agent:
    return get_agent('agent4')


@pytest.fixture()
def A() -> Agent:
    return get_agent('agent1')


@pytest.fixture()
def B() -> Agent:
    return get_agent('agent2')


@pytest.fixture()
def C() -> Agent:
    return get_agent('agent3')


@pytest.fixture()
def D() -> Agent:
    return get_agent('agent4')


@pytest.fixture()
def ledger_name() -> str:
    return 'Ledger-' + uuid.uuid4().hex


@pytest.fixture()
def ledger_names() -> list:
    return ['Ledger-' + uuid.uuid4().hex for n in range(2)]


@pytest.fixture()
def default_network() -> str:
    return 'default'


async def get_pairwise(me: Agent, their: Agent):
    suite = get_suite_singleton()
    me_params = suite.get_agent_params(me.name)
    their_params = suite.get_agent_params(their.name)
    me_label, me_entity = list(me_params['entities'].keys())[0], list(me_params['entities'].items())[0][1]
    their_label, their_entity = list(their_params['entities'].keys())[0], list(their_params['entities'].items())[0][1]
    me_endpoint_address = [e for e in me.endpoints if e.routing_keys == []][0].address
    their_endpoint_address = [e for e in their.endpoints if e.routing_keys == []][0].address
    self = me
    for agent, entity_me, entity_their, label_their, endpoint_their in [
        (me, me_entity, their_entity, their_label, their_endpoint_address),
        (their, their_entity, me_entity, me_label, me_endpoint_address)
    ]:
        pairwise = await agent.pairwise_list.load_for_did(their_did=their_entity['did'])
        is_filled = pairwise and pairwise.metadata
        if not is_filled:
            me_ = Pairwise.Me(entity_me['did'], entity_me['verkey'])
            their_ = Pairwise.Their(entity_their['did'], their_label, endpoint_their, entity_their['verkey'])
            metadata = {
                'me': {
                    'did': entity_me['did'],
                    'verkey': entity_me['verkey'],
                    'did_doc': None
                },
                'their': {
                    'did': entity_their['did'],
                    'verkey': entity_their['verkey'],
                    'label': label_their,
                    'endpoint': {
                        'address': endpoint_their,
                        'routing_keys': []
                    },
                    'did_doc': None
                }
            }
            pairwise = Pairwise(me=me_, their=their_, metadata=metadata)
            await agent.wallet.did.store_their_did(entity_their['did'], entity_their['verkey'])
            await agent.pairwise_list.ensure_exists(pairwise)
    pairwise = await self.pairwise_list.load_for_did(their_did=their_entity['did'])
    return pairwise



from abc import ABC, abstractmethod
from typing import Any, List


class AbstractImmutableCollection(ABC):

    @abstractmethod
    async def select_db(self, db_name: str):
        raise NotImplemented

    @abstractmethod
    async def add(self, value: Any, tags: dict):
        raise NotImplemented

    @abstractmethod
    async def fetch(self, tags: dict, limit: int = None) -> (List[Any], int):
        raise NotImplemented



import contextlib
from abc import abstractmethod
from typing import Union
from datetime import datetime, timedelta

import sirius_sdk
from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.hub import CoProtocolP2P
from sirius_sdk.agent.codec import encode
from sirius_sdk.agent.aries_rfc.utils import utc_to_str
from sirius_sdk.agent.ledger import Schema, CredentialDefinition
from sirius_sdk.errors.indy_exceptions import WalletItemNotFound
from sirius_sdk.base import AbstractStateMachine
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential.messages import *


PROPOSE_NOT_ACCEPTED = "propose_not_accepted"
OFFER_PROCESSING_ERROR = 'offer_processing_error'
REQUEST_NOT_ACCEPTED = "request_not_accepted"
ISSUE_PROCESSING_ERROR = 'issue_processing_error'
RESPONSE_FOR_UNKNOWN_REQUEST = "response_for_unknown_request"


class BaseIssuingStateMachine(AbstractStateMachine):

    def __init__(self, time_to_live: int = 60, logger=None, *args, **kwargs):
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self._problem_report = None
        self.__time_to_live = time_to_live
        self.__coprotocol: Optional[CoProtocolP2P] = None

    @property
    def time_to_live(self) -> Optional[int]:
        return self.__time_to_live

    @property
    def problem_report(self) -> IssueProblemReport:
        return self._problem_report

    @contextlib.asynccontextmanager
    async def coprotocol(self, pairwise: Pairwise):
        self.__coprotocol = CoProtocolP2P(
            pairwise=pairwise,
            protocols=[BaseIssueCredentialMessage.PROTOCOL, Ack.PROTOCOL],
            time_to_live=self.time_to_live
        )
        self._register_for_aborting(self.__coprotocol)
        try:
            try:
                yield
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            self._unregister_for_aborting(self.__coprotocol)

    async def switch(self, request: BaseIssueCredentialMessage) -> Union[BaseIssueCredentialMessage, Ack]:
        ok, resp = await self.__coprotocol.switch(request)
        if ok:
            if isinstance(resp, BaseIssueCredentialMessage) or isinstance(resp, Ack):
                try:
                    resp.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(
                        ISSUE_PROCESSING_ERROR if self._is_leader() else REQUEST_NOT_ACCEPTED,
                        e.message
                    )
                return resp
            elif isinstance(resp, IssueProblemReport):
                raise StateMachineTerminatedWithError(resp.problem_code, resp.explain, notify=False)
            else:
                raise StateMachineTerminatedWithError(
                    ISSUE_PROCESSING_ERROR if self._is_leader() else REQUEST_NOT_ACCEPTED,
                    'Unexpected response @type: %s' % str(resp.type)
                )
        else:
            raise StateMachineTerminatedWithError(
                ISSUE_PROCESSING_ERROR if self._is_leader() else REQUEST_NOT_ACCEPTED,
                'Response awaiting terminated by timeout'
            )

    async def send(self, msg: Union[BaseIssueCredentialMessage, Ack, IssueProblemReport]):
        await self.__coprotocol.send(msg)

    @abstractmethod
    def _is_leader(self) -> bool:
        raise NotImplemented


class Issuer(BaseIssuingStateMachine):
    """Implementation of Issuer role for Credential-issuing protocol

    See details: https://github.com/hyperledger/aries-rfcs/tree/master/features/0036-issue-credential
    """

    def __init__(self, holder: Pairwise, time_to_live: int = 60, logger=None, *args, **kwargs):
        """
        :param holder: Holder side described as pairwise instance.
          (Assumed pairwise was established earlier: statically or via connection-protocol)
        """

        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__holder = holder

    async def issue(
            self, values: dict, schema: Schema, cred_def: CredentialDefinition,
            comment: str = None, locale: str = BaseIssueCredentialMessage.DEF_LOCALE,
            preview: List[ProposedAttrib] = None, translation: List[AttribTranslation] = None, cred_id: str = None
    ) -> bool:
        """
        :param values: credential values {"attr_name": "attr_value"}
        :param schema: credential schema
        :param cred_def: credential definition prepared and stored in Ledger earlier
        :param comment: human readable credential comment
        :param preview: credential preview
        :param locale: locale, for example "en" or "ru"
        :param translation: translation of the credential preview according to locale
        :param cred_id: credential id. Issuer may issue multiple credentials with same cred-id to give holder ability
                        to restore old credential
        """
        async with self.coprotocol(pairwise=self.__holder):
            try:
                # Step-1: Send offer to holder
                offer = await sirius_sdk.AnonCreds.issuer_create_credential_offer(cred_def_id=cred_def.id)
                expires_time = datetime.utcnow() + timedelta(seconds=self.time_to_live)
                offer_msg = OfferCredentialMessage(
                    comment=comment,
                    locale=locale,
                    offer=offer,
                    cred_def=cred_def.body,
                    preview=preview,
                    issuer_schema=schema.body,
                    translation=translation,
                    expires_time=utc_to_str(expires_time)
                )
                await self.log(progress=20, message='Send offer', payload=dict(offer_msg))

                # Switch to await participant action
                resp = await self.switch(offer_msg)
                if not isinstance(resp, RequestCredentialMessage):
                    raise StateMachineTerminatedWithError(
                        OFFER_PROCESSING_ERROR, 'Unexpected @type: %s' % str(resp.type)
                    )

                # Step-2: Create credential
                request_msg = resp
                await self.log(progress=40, message='Received credential request', payload=dict(request_msg))
                encoded_cred_values = dict()
                for key, value in values.items():
                    encoded_cred_values[key] = dict(raw=str(value), encoded=encode(value))
                await self.log(progress=70, message='Build credential with values', payload=encoded_cred_values)

                ret = await sirius_sdk.AnonCreds.issuer_create_credential(
                    cred_offer=offer,
                    cred_req=request_msg.cred_request,
                    cred_values=encoded_cred_values,
                    rev_reg_id=None,
                    blob_storage_reader_handle=None
                )
                cred, cred_revoc_id, revoc_reg_delta = ret

                # Step-3: Issue and wait Ack
                issue_msg = IssueCredentialMessage(
                    comment=comment,
                    locale=locale,
                    cred=cred,
                    cred_id=cred_id
                )
                await self.log(progress=90, message='Send Issue message', payload=dict(issue_msg))

                ack = await self.switch(issue_msg)
                if not isinstance(ack, Ack):
                    raise StateMachineTerminatedWithError(
                        ISSUE_PROCESSING_ERROR, 'Unexpected @type: %s' % str(resp.type)
                    )
                await self.log(progress=100, message='Issuing was terminated successfully')
                return True
            except StateMachineTerminatedWithError as e:
                self._problem_report = IssueProblemReport(
                    problem_code=e.problem_code,
                    explain=e.explain,
                )
                if e.notify:
                    await self.send(self._problem_report)
                await self.log(
                    progress=100, message=f'Terminated with error',
                    problem_code=e.problem_code, explain=e.explain
                )
                return False

    def _is_leader(self) -> bool:
        return True


class Holder(BaseIssuingStateMachine):
    """Implementation of Holder role for Credential-issuing protocol

    See details: https://github.com/hyperledger/aries-rfcs/tree/master/features/0036-issue-credential
    """

    def __init__(
            self, issuer: Pairwise, time_to_live: int = 60, logger=None, *args, **kwargs
    ):
        """
        :param issuer: Issuer described as pairwise instance.
          (Assumed pairwise was established earlier: statically or via connection-protocol)
        """
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__issuer = issuer

    async def accept(
            self, offer: OfferCredentialMessage, master_secret_id: str,
            comment: str = None, locale: str = BaseIssueCredentialMessage.DEF_LOCALE,
    ) -> (bool, Optional[str]):
        """
        :param offer: credential offer
        :param master_secret_id: prover master secret ID
        :param comment: human readable comment
        :param locale: locale, for example "en" or "ru"
        """
        doc_uri = offer.doc_uri
        async with self.coprotocol(pairwise=self.__issuer):
            try:
                offer_msg = offer
                try:
                    offer_msg.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)

                # Step-1: Process Issuer Offer
                cred_request, cred_metadata = await sirius_sdk.AnonCreds.prover_create_credential_req(
                    prover_did=self.__issuer.me.did,
                    cred_offer=offer_msg.offer,
                    cred_def=offer_msg.cred_def,
                    master_secret_id=master_secret_id
                )

                # Step-2: Send request to Issuer
                request_msg = RequestCredentialMessage(
                    comment=comment,
                    locale=locale,
                    cred_request=cred_request,
                    doc_uri=doc_uri
                )

                # Switch to await participant action
                resp = await self.switch(request_msg)
                if not isinstance(resp, IssueCredentialMessage):
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, 'Unexpected @type: %s' % str(resp.type))

                issue_msg = resp
                try:
                    issue_msg.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)

                # Step-3: Store credential
                cred_id = await self._store_credential(
                    cred_metadata, issue_msg.cred, offer.cred_def, None, issue_msg.cred_id
                )
                ack = Ack(
                    thread_id=issue_msg.ack_message_id if issue_msg.please_ack else issue_msg.id,
                    status=Status.OK,
                    doc_uri=doc_uri
                )
                await self.send(ack)

            except StateMachineTerminatedWithError as e:
                self._problem_report = IssueProblemReport(
                    problem_code=e.problem_code,
                    explain=e.explain,
                    doc_uri=doc_uri
                )
                await self.log(
                    progress=100, message=f'Terminated with error',
                    problem_code=e.problem_code, explain=e.explain
                )
                if e.notify:
                    await self.send(self._problem_report)
                return False, None
            else:
                return True, cred_id

    def _is_leader(self) -> bool:
        return False

    @staticmethod
    async def _store_credential(
            cred_metadata: dict, cred: dict, cred_def: dict, rev_reg_def: Optional[dict], cred_id: Optional[str]
    ) -> str:
        try:
            cred_older = await sirius_sdk.AnonCreds.prover_get_credential(cred_id)
        except WalletItemNotFound:
            cred_older = None
        if cred_older:
            # Delete older credential
            await sirius_sdk.AnonCreds.prover_delete_credential(cred_id)
        cred_id = await sirius_sdk.AnonCreds.prover_store_credential(
            cred_req_metadata=cred_metadata,
            cred=cred,
            cred_def=cred_def,
            rev_reg_def=rev_reg_def,
            cred_id=cred_id
        )
        return cred_id



from typing import List, Optional

from sirius_sdk.agent.wallet import RetrieveRecordOptions
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.agent.connections import AgentRPC


class NonSecretsProxy(AbstractNonSecrets):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def add_wallet_record(self, type_: str, id_: str, value: str, tags: dict = None) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/add_wallet_record',
            params=dict(type_=type_, id_=id_, value=value, tags=tags)
        )

    async def update_wallet_record_value(self, type_: str, id_: str, value: str) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/update_wallet_record_value',
            params=dict(type_=type_, id_=id_, value=value)
        )

    async def update_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/update_wallet_record_tags',
            params=dict(type_=type_, id_=id_, tags=tags)
        )

    async def add_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/add_wallet_record_tags',
            params=dict(type_=type_, id_=id_, tags=tags)
        )

    async def delete_wallet_record_tags(self, type_: str, id_: str, tag_names: List[str]) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/delete_wallet_record_tags',
            params=dict(type_=type_, id_=id_, tag_names=tag_names)
        )

    async def delete_wallet_record(self, type_: str, id_: str) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/delete_wallet_record',
            params=dict(type_=type_, id_=id_)
        )

    async def get_wallet_record(self, type_: str, id_: str, options: RetrieveRecordOptions) -> Optional[dict]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_wallet_record',
            params=dict(type_=type_, id_=id_, options=options)
        )

    async def wallet_search(
            self, type_: str, query: dict, options: RetrieveRecordOptions, limit: int = 1
    ) -> (List[dict], int):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/wallet_search',
            params=dict(type_=type_, query=query, options=options, limit=limit)
        )



from functools import lru_cache
from typing import Optional, Union
import base64

import base58
import nacl.bindings
import nacl.exceptions
import nacl.utils

from ..errors.exceptions import SiriusCryptoError


def b64_to_bytes(value: Union[str, bytes], urlsafe: bool=False) -> bytes:
    """Convert a base 64 string to bytes.

    :param value: (str, bytes) input base64 value
    :param urlsafe: (bool) flag if needed to convert to urlsafe presentation
    :return: bytes array
    """
    if isinstance(value, str):
        value = value.encode('ascii')
    if urlsafe:
        missing_padding = len(value) % 4
        if missing_padding:
            value += b'=' * (4 - missing_padding)
        return base64.urlsafe_b64decode(value)
    return base64.b64decode(value)


def bytes_to_b64(value: bytes, urlsafe=False) -> str:
    """Convert a byte string to base 64.

    :param value: (bytes) input bytes array
    :param urlsafe: (bool) flag if needed to convert to urlsafe presentation
    :return base64 presentation
    """
    if urlsafe:
        return base64.urlsafe_b64encode(value).decode("ascii")
    else:
        return base64.b64encode(value).decode("ascii")


@lru_cache(maxsize=16)
def b58_to_bytes(value: str) -> bytes:
    """
    Convert a base 58 string to bytes.

    Small cache provided for key conversions which happen frequently in pack
    and unpack and message handling.
    """
    return base58.b58decode(value)


@lru_cache(maxsize=16)
def bytes_to_b58(value: bytes) -> str:
    """
    Convert a byte string to base 58.

    Small cache provided for key conversions which happen frequently in pack
    and unpack and message handling.
    """
    return base58.b58encode(value).decode("ascii")


def create_keypair(seed: bytes = None) -> (bytes, bytes):
    """
    Create a public and private signing keypair from a seed value.

    :param seed: (bytes) Seed for keypair
    :return A tuple of (public key, secret key)
    """
    if seed:
        validate_seed(seed)
    else:
        seed = random_seed()
    pk, sk = nacl.bindings.crypto_sign_seed_keypair(seed)
    return pk, sk


def random_seed() -> bytes:
    """
    Generate a random seed value.

    :return A new random seed
    """
    return nacl.utils.random(nacl.bindings.crypto_secretbox_KEYBYTES)


def validate_seed(seed: Union[str, bytes]) -> Optional[bytes]:
    """
    Convert a seed parameter to standard format and check length.

    :param seed: (str, bytes) The seed to validate
    :return The validated and encoded seed
    """
    if not seed:
        return None
    if isinstance(seed, str):
        if "=" in seed:
            seed = b64_to_bytes(seed)
        else:
            seed = seed.encode("ascii")
    if not isinstance(seed, bytes):
        raise SiriusCryptoError("Seed value is not a string or bytes")
    if len(seed) != 32:
        raise SiriusCryptoError("Seed value must be 32 bytes in length")
    return seed



import sys
import asyncio
from typing import Optional, List

from sirius_sdk.agent.connections import AgentEvents
from sirius_sdk.errors.exceptions import SiriusConnectionClosed
from sirius_sdk.messaging import Message, restore_message_instance
from sirius_sdk.agent.pairwise import AbstractPairwiseList, Pairwise


PY_35 = sys.version_info >= (3, 5)
PY_352 = sys.version_info >= (3, 5, 2)


class Event(Message):
    
    def __init__(self, pairwise: Pairwise = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__pairwise = pairwise

    @property
    def pairwise(self) -> Pairwise:
        return self.__pairwise

    @property
    def message(self) -> Optional[Message]:
        if 'message' in self:
            return self['message']
        else:
            return None

    @property
    def recipient_verkey(self) -> Optional[str]:
        return self.get('recipient_verkey', None)

    @property
    def sender_verkey(self) -> Optional[str]:
        return self.get('sender_verkey', None)

    @property
    def forwarded_keys(self) -> List[str]:
        return self.get('forwarded_keys', [])

    @property
    def content_type(self) -> Optional[str]:
        return self.get('content_type', None)

    @property
    def extra(self) -> Optional[dict]:
        return self.get('~extra', None)


class Listener:

    def __init__(self, source: AgentEvents, pairwise_resolver: AbstractPairwiseList = None):
        self.__source = source
        self.__pairwise_resolver = pairwise_resolver

    async def get_one(self, timeout: int = None) -> Event:
        event = await self.__source.pull(timeout)
        if 'message' in event:
            ok, message = restore_message_instance(event['message'])
            if ok:
                event['message'] = message
            else:
                event['message'] = Message(event['message'])
        their_verkey = event.get('sender_verkey', None)
        if self.__pairwise_resolver and their_verkey:
            pairwise = await self.__pairwise_resolver.load_for_verkey(their_verkey)
        else:
            pairwise = None
        return Event(pairwise=pairwise, **event)

    if PY_35:
        def __aiter__(self):
            if not self.__source.is_open:
                raise SiriusConnectionClosed()
            return self

        # Old 3.5 versions require a coroutine
        if not PY_352:
            __aiter__ = asyncio.coroutine(__aiter__)

        @asyncio.coroutine
        def __anext__(self):
            """Asyncio iterator interface for listener

            Note:
                TopicAuthorizationFailedError and OffsetOutOfRangeError
                exceptions can be raised in iterator.
                All other KafkaError exceptions will be logged and not raised
            """
            while True:
                return (yield from self.get_one())



import uuid
import json
from typing import List
from datetime import datetime

import pytest

import sirius_sdk
from sirius_sdk import Agent, Pairwise
from sirius_sdk.agent.codec import encode
from sirius_sdk.agent.aries_rfc.utils import str_to_utc
from sirius_sdk.agent.ledger import CredentialDefinition
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.state_machines import Verifier, Prover, \
    AttribTranslation, RequestPresentationMessage
from sirius_sdk.errors.indy_exceptions import AnoncredsMasterSecretDuplicateNameError

from .conftest import get_pairwise
from .helpers import run_coroutines, IndyAgent, ServerTestSuite


async def run_verifier(
        uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        prover: Pairwise, proof_request: dict, translation: List[AttribTranslation] = None
) -> bool:
    async with sirius_sdk.context(uri, credentials, p2p):
        ledger = await sirius_sdk.ledger('default')
        machine = Verifier(prover=prover, ledger=ledger)
        success = await machine.verify(
            proof_request, translation=translation, comment='I am Verifier', proto_version='1.0'
        )
        if not success:
            print('===================== Verifier terminated with error ====================')
            if machine.problem_report:
                print(json.dumps(machine.problem_report, indent=2, sort_keys=True))
            print('=======================================================================')
        return success


async def run_prover(
        uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        verifier: Pairwise, master_secret_id: str
):
    async with sirius_sdk.context(uri, credentials, p2p):
        listener = await sirius_sdk.subscribe()
        event = await listener.get_one()
        assert event.pairwise is not None
        assert event.pairwise.their.did == verifier.their.did
        request = event.message
        assert isinstance(request, RequestPresentationMessage)
        ttl = 60
        if request.expires_time:
            expire = str_to_utc(request.expires_time, raise_exceptions=False)
            delta = expire - datetime.utcnow()
            if delta.seconds > 0:
                ttl = delta.seconds
        try:
            ledger = await sirius_sdk.ledger('default')
            machine = Prover(verifier=verifier, ledger=ledger, time_to_live=ttl)
            success = await machine.prove(request, master_secret_id)
            if not success:
                print('===================== Prover terminated with error ====================')
                if machine.problem_report:
                    print(json.dumps(machine.problem_report, indent=2, sort_keys=True))
                print('=======================================================================')
            return success
        except Exception as e:
            print('==== Prover routine Exception: ' + repr(e))


@pytest.mark.asyncio
async def test_sane(
        test_suite: ServerTestSuite, agent1: Agent, agent2: Agent, agent3: Agent, prover_master_secret_name: str
):
    issuer = agent1
    prover = agent2
    verifier = agent3
    await issuer.open()
    await prover.open()
    await verifier.open()
    try:
        print('Establish pairwises')
        i2p = await get_pairwise(issuer, prover)
        p2i = await get_pairwise(prover, issuer)
        v2p = await get_pairwise(verifier, prover)
        p2v = await get_pairwise(prover, verifier)

        print('Register schema')
        did_issuer, verkey_issuer = i2p.me.did, i2p.me.verkey
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        print('Register credential def')
        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True

        print('Prepare Prover')
        try:
            await prover.wallet.anoncreds.prover_create_master_secret(prover_master_secret_name)
        except AnoncredsMasterSecretDuplicateNameError:
            pass

        prover_secret_id = prover_master_secret_name
        cred_values = {'attr1': 'Value-1', 'attr2': 456, 'attr3': 5.87}
        cred_id = 'cred-id-' + uuid.uuid4().hex

        # Issue credential
        offer = await issuer.wallet.anoncreds.issuer_create_credential_offer(cred_def_id=cred_def.id)
        cred_request, cred_metadata = await prover.wallet.anoncreds.prover_create_credential_req(
            prover_did=p2i.me.did, cred_offer=offer, cred_def=cred_def.body, master_secret_id=prover_secret_id
        )
        encoded_cred_values = dict()
        for key, value in cred_values.items():
            encoded_cred_values[key] = dict(raw=str(value), encoded=encode(value))
        ret = await issuer.wallet.anoncreds.issuer_create_credential(
            cred_offer=offer,
            cred_req=cred_request,
            cred_values=encoded_cred_values,
            rev_reg_id=None,
            blob_storage_reader_handle=None
        )
        cred, cred_revoc_id, revoc_reg_delta = ret
        await prover.wallet.anoncreds.prover_store_credential(
            cred_req_metadata=cred_metadata,
            cred=cred,
            cred_def=cred_def.body,
            rev_reg_def=None,
            cred_id=cred_id
        )

    finally:
        await issuer.close()
        await prover.close()
        await verifier.close()

    prover = test_suite.get_agent_params('agent2')
    verifier = test_suite.get_agent_params('agent3')

    # FIRE !!!
    attr_referent_id = 'attr1_referent'
    pred_referent_id = 'predicate1_referent'
    async with sirius_sdk.context(verifier['server_address'], verifier['credentials'], verifier['p2p']):
        proof_request = {
            "nonce": await sirius_sdk.AnonCreds.generate_nonce(),
            "name": "Test ProofRequest",
            "version": "0.1",
            "requested_attributes": {
                attr_referent_id: {
                    "name": "attr1",
                    "restrictions": {
                        "issuer_did": did_issuer
                    }
                }
            },
            "requested_predicates": {
                pred_referent_id: {
                    'name': 'attr2',
                    'p_type': '>=',
                    'p_value': 100,
                    "restrictions": {
                        "issuer_did": did_issuer
                    }
                }
            }
        }

    coro_verifier = run_verifier(
        verifier['server_address'], verifier['credentials'], verifier['p2p'],
        prover=v2p,
        proof_request=proof_request
    )
    coro_prover = run_prover(
        prover['server_address'], prover['credentials'], prover['p2p'],
        verifier=p2v,
        master_secret_id=prover_secret_id
    )
    print('Run state machines')
    results = await run_coroutines(coro_verifier, coro_prover, timeout=60)
    print('Finish state machines')
    print(str(results))
    assert len(results) == 2
    for res in results:
        assert res is True


@pytest.mark.asyncio
async def test_multiple_provers(
    test_suite: ServerTestSuite,
    agent1: Agent, agent2: Agent, agent3: Agent, agent4: Agent, prover_master_secret_name: str
):
    issuer = agent1
    prover1 = agent2
    verifier = agent3
    prover2 = agent4
    await issuer.open()
    await prover1.open()
    await prover2.open()
    await verifier.open()
    try:
        print('Establish pairwises')
        i_to_p1 = await get_pairwise(issuer, prover1)
        i_to_p2 = await get_pairwise(issuer, prover2)
        p1_to_i = await get_pairwise(prover1, issuer)
        p2_to_i = await get_pairwise(prover2, issuer)
        v_to_p1 = await get_pairwise(verifier, prover1)
        v_to_p2 = await get_pairwise(verifier, prover2)
        p1_to_v = await get_pairwise(prover1, verifier)
        p2_to_v = await get_pairwise(prover2, verifier)

        print('Register schema')
        did_issuer, verkey_issuer = i_to_p1.me.did, i_to_p1.me.verkey
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        print('Register credential def')
        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True

        print('Prepare Provers')
        for prover in [prover1, prover2]:
            try:
                await prover.wallet.anoncreds.prover_create_master_secret(prover_master_secret_name)
            except AnoncredsMasterSecretDuplicateNameError:
                pass

        cred_ids = {
            0: 'cred-id-' + uuid.uuid4().hex,
            1: 'cred-id-' + uuid.uuid4().hex
        }
        prover_did = {
            0: p1_to_i.me.did,
            1: p2_to_i.me.did
        }
        for i, prover in enumerate([prover1, prover2]):
            prover_secret_id = prover_master_secret_name
            cred_values = {'attr1': f'Value-{i}', 'attr2': 200 + i*10, 'attr3': i*1.5}
            cred_id = cred_ids[0]

            # Issue credential
            offer = await issuer.wallet.anoncreds.issuer_create_credential_offer(cred_def_id=cred_def.id)
            cred_request, cred_metadata = await prover.wallet.anoncreds.prover_create_credential_req(
                prover_did=prover_did[i], cred_offer=offer, cred_def=cred_def.body, master_secret_id=prover_secret_id
            )
            encoded_cred_values = dict()
            for key, value in cred_values.items():
                encoded_cred_values[key] = dict(raw=str(value), encoded=encode(value))
            ret = await issuer.wallet.anoncreds.issuer_create_credential(
                cred_offer=offer,
                cred_req=cred_request,
                cred_values=encoded_cred_values,
                rev_reg_id=None,
                blob_storage_reader_handle=None
            )
            cred, cred_revoc_id, revoc_reg_delta = ret
            await prover.wallet.anoncreds.prover_store_credential(
                cred_req_metadata=cred_metadata,
                cred=cred,
                cred_def=cred_def.body,
                rev_reg_def=None,
                cred_id=cred_id
            )

    finally:
        await issuer.close()
        await prover1.close()
        await prover2.close()
        await verifier.close()

    prover1 = test_suite.get_agent_params('agent2')
    verifier = test_suite.get_agent_params('agent3')
    prover2 = test_suite.get_agent_params('agent4')
    # FIRE !!!
    attr_referent_id = 'attr1_referent'
    pred_referent_id = 'predicate1_referent'
    async with sirius_sdk.context(verifier['server_address'], verifier['credentials'], verifier['p2p']):
        proof_request = {
            "nonce": await sirius_sdk.AnonCreds.generate_nonce(),
            "name": "Test ProofRequest",
            "version": "0.1",
            "requested_attributes": {
                attr_referent_id: {
                    "name": "attr1",
                    "restrictions": {
                        "issuer_did": did_issuer
                    }
                }
            },
            "requested_predicates": {
                pred_referent_id: {
                    'name': 'attr2',
                    'p_type': '>=',
                    'p_value': 100,
                    "restrictions": {
                        "issuer_did": did_issuer
                    }
                }
            }
        }

    for prover, v2p, p2v in [(prover1, v_to_p1, p1_to_v), (prover2, v_to_p2, p2_to_v)]:
        coro_verifier = run_verifier(
            verifier['server_address'], verifier['credentials'], verifier['p2p'],
            prover=v2p,
            proof_request=proof_request
        )
        coro_prover = run_prover(
            prover['server_address'], prover['credentials'], prover['p2p'],
            verifier=p2v,
            master_secret_id=prover_master_secret_name
        )
        print('Run state machines')
        results = await run_coroutines(coro_verifier, coro_prover, timeout=60)
        print('Finish state machines')
        print(str(results))
        assert len(results) == 2
        for res in results:
            assert res is True



from sirius_sdk.rpc.futures import Future
from sirius_sdk.rpc.parsing import build_request
from sirius_sdk.rpc.tunnel import AddressedTunnel


__all__ = ["Future", "build_request", "AddressedTunnel"]



import json
import hashlib
from typing import List, Optional, Union

from sirius_sdk.encryption import bytes_to_b58
from sirius_sdk.errors.exceptions import *
from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.agent.microledgers.abstract import serialize_ordering
from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, AriesProblemReport, THREAD_DECORATOR
from sirius_sdk.agent.microledgers.abstract import Transaction, AbstractMicroledger
from sirius_sdk.agent.aries_rfc.utils import sign, verify_signed


class SimpleConsensusMessage(AriesProtocolMessage, metaclass=RegisterMessage):
    """Message for Simple Consensus protocol over Microledger maintenance

    """
    PROTOCOL = 'simple-consensus'

    def __init__(self, participants: List[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['participants'] = participants or []

    @property
    def participants(self) -> List[str]:
        return self.get('participants', [])


class SimpleConsensusProblemReport(AriesProblemReport, metaclass=RegisterMessage):
    PROTOCOL = SimpleConsensusMessage.PROTOCOL


class BaseInitLedgerMessage(SimpleConsensusMessage):
    NAME = 'initialize'

    def __init__(
            self, ledger_name: Optional[str] = None, genesis: List[Transaction] = None,
            root_hash: Optional[str] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        ledger = {}
        if ledger_name is not None:
            ledger['name'] = ledger_name
        if root_hash is not None:
            ledger['root_hash'] = root_hash
        if genesis is not None:
            ledger['genesis'] = genesis
        if ledger:
            self['ledger'] = ledger
            hashfunc = hashlib.sha256
            hasher = hashfunc()
            data = serialize_ordering(ledger)
            hasher.update(data)
            digest = hasher.digest()
            self['ledger~hash'] = {
                'func': 'sha256',
                'base58': bytes_to_b58(digest)
            }

    @property
    def ledger(self) -> Optional[dict]:
        return self.get('ledger', None)

    @property
    def ledger_hash(self) -> Optional[dict]:
        return self.get('ledger~hash', None)

    @property
    def signatures(self) -> List[dict]:
        return self.get('signatures', [])

    async def check_signatures(self, api: AbstractCrypto, participant: str = 'ALL') -> dict:
        if self.ledger_hash is None:
            raise SiriusContextError('Ledger Hash description is empty')
        if participant == 'ALL':
            signatures = self.signatures
        else:
            signatures = [s for s in self.signatures if s['participant'] == participant]
        if signatures:
            response = {}
            for item in signatures:
                signed_ledger_hash, is_success = await verify_signed(api, item['signature'])
                if not is_success:
                    raise SiriusValidationError('Invalid Sign for participant: "%s"' % item['participant'])
                if signed_ledger_hash != self.ledger_hash:
                    raise SiriusValidationError('NonConsistent Ledger hash for participant: "%s"' % item['participant'])
                response[item['participant']] = signed_ledger_hash
            return response
        else:
            raise SiriusContextError('Signatures list is empty!')


class InitRequestLedgerMessage(BaseInitLedgerMessage):

    NAME = 'initialize-request'

    def __init__(self, timeout_sec: int = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if timeout_sec:
            self['timeout_sec'] = timeout_sec

    @property
    def timeout_sec(self) -> Optional[int]:
        return self.get('timeout_sec', None)

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    async def add_signature(self, api: AbstractCrypto, me: Pairwise.Me):
        if me.did not in self.participants:
            raise SiriusContextError('Signer must be a participant')
        if self.ledger_hash is not None:
            hash_signature = await sign(api, self.ledger_hash, me.verkey)
            signatures = [s for s in self.signatures if s['participant'] != me.did]
            signatures.append(
                {
                    'participant': me.did,
                    'signature': hash_signature
                }
            )
            self['signatures'] = signatures
        else:
            raise SiriusContextError('Ledger Hash description is empty')

    async def check_ledger_hash(self):
        if not self.ledger_hash:
            raise SiriusContextError('Ledger hash is empty')
        if not self.ledger:
            raise SiriusContextError('Ledger body is empty')

    def validate(self):
        super().validate()
        if not self.ledger:
            raise SiriusValidationError('Ledger info is empty')
        for expect_field in ['root_hash', 'name', 'genesis']:
            if expect_field not in self.ledger.keys():
                raise SiriusValidationError(f'Expected field "{expect_field}" does not exists in Ledger container')
        if not self.ledger_hash:
            raise SiriusValidationError('Ledger Hash info is empty')
        for expect_field in ['func', 'base58']:
            if expect_field not in self.ledger_hash.keys():
                raise SiriusValidationError(f'Expected field "{expect_field}" does not exists in Ledger Hash')


class InitResponseLedgerMessage(InitRequestLedgerMessage):

    NAME = 'initialize-response'

    def assign_from(self, source: BaseInitLedgerMessage):
        partial = {k: v for k, v in source.items() if k not in ['@id', '@type', THREAD_DECORATOR]}
        self.update(partial)

    def signature(self, did: str) -> Optional[dict]:
        filtered = [p for p in self.signatures if p['participant'] == did]
        return filtered[0] if filtered else None


class MicroLedgerState(dict):

    @classmethod
    def from_ledger(cls, ledger: AbstractMicroledger):
        return MicroLedgerState(
            {
                'name': ledger.name,
                'seq_no': ledger.seq_no,
                'size': ledger.size,
                'uncommitted_size': ledger.uncommitted_size,
                'root_hash': ledger.root_hash,
                'uncommitted_root_hash': ledger.uncommitted_root_hash
            }
        )

    def is_filled(self) -> bool:
        return all(
            [
                k in self.keys() for k in
                ('name', 'seq_no', 'size', 'uncommitted_size', 'root_hash', 'uncommitted_root_hash')
            ]
        )

    @property
    def name(self) -> str:
        return self['name']

    @name.setter
    def name(self, value: int):
        self['name'] = value

    @property
    def seq_no(self) -> int:
        return self['seq_no']

    @seq_no.setter
    def seq_no(self, value: int):
        self['seq_no'] = value

    @property
    def size(self) -> int:
        return self['size']

    @size.setter
    def size(self, value: int):
        self['size'] = value

    @property
    def uncommitted_size(self) -> int:
        return self['uncommitted_size']

    @uncommitted_size.setter
    def uncommitted_size(self, value: int):
        self['uncommitted_size'] = value

    @property
    def root_hash(self) -> str:
        return self['root_hash']

    @root_hash.setter
    def root_hash(self, value: str):
        self['root_hash'] = value

    @property
    def uncommitted_root_hash(self) -> str:
        return self['uncommitted_root_hash']

    @uncommitted_root_hash.setter
    def uncommitted_root_hash(self, value: str):
        self['uncommitted_root_hash'] = value

    @property
    def hash(self) -> str:
        dump = json.dumps(self, sort_keys=True, ensure_ascii=False, separators=(',', ':'))
        return hashlib.md5(dump.encode()).hexdigest()


class BaseTransactionsMessage(SimpleConsensusMessage):

    NAME = 'stage'

    def __init__(self, transactions: List[Transaction] = None, state: Optional[MicroLedgerState] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if transactions is not None:
            for txn in transactions:
                txn = Transaction(txn)
                if not txn.has_metadata():
                    raise SiriusContextError('Transaction must have processed by Ledger engine and has metadata')
            self['transactions'] = transactions
        if state:
            state = MicroLedgerState(state)
            self['state'] = state
            self['hash'] = state.hash

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @property
    def transactions(self) -> Optional[List[Transaction]]:
        txns = self.get('transactions', None)
        if txns is not None:
            return [Transaction(txn) for txn in txns]
        else:
            return None

    @property
    def state(self) -> Optional[MicroLedgerState]:
        state = self.get('state', None)
        if state is not None:
            state = MicroLedgerState(state)
            return state if state.is_filled() else None
        else:
            return None

    @property
    def hash(self) -> Optional[str]:
        return self.get('hash', None)


class ProposeTransactionsMessage(BaseTransactionsMessage):
    """Message to process transactions propose by Actor
    """
    NAME = 'stage-propose'

    def __init__(self, timeout_sec: int = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if timeout_sec:
            self['timeout_sec'] = timeout_sec

    @property
    def timeout_sec(self) -> Optional[int]:
        return self.get('timeout_sec', None)

    def validate(self):
        super().validate()
        if not self.transactions:
            raise SiriusValidationError('Empty transactions list')
        for txn in self.transactions:
            if not txn.has_metadata():
                raise SiriusValidationError('Transaction has not metadata')
        if not self.state:
            raise SiriusValidationError('Empty state')
        if not self.hash:
            raise SiriusValidationError('Empty hash')


class PreCommitTransactionsMessage(BaseTransactionsMessage):
    """Message to accumulate participants signed accepts for transactions list
    """
    NAME = 'stage-pre-commit'

    async def sign_state(self, api: AbstractCrypto, me: Pairwise.Me):
        signed = await sign(api, self.hash, me.verkey)
        self['hash~sig'] = signed
        del self['state']

    async def verify_state(self, api: AbstractCrypto, expected_verkey: str) -> (bool, Optional[str]):
        hash_signed = self.get('hash~sig', None)
        if hash_signed:
            if hash_signed['signer'] == expected_verkey:
                state_hash, is_success = await verify_signed(api, hash_signed)
                return is_success, state_hash
            else:
                return False, None
        else:
            return False, None


class CommitTransactionsMessage(BaseTransactionsMessage):
    """Message to commit transactions list
    """
    NAME = 'stage-commit'

    @property
    def pre_commits(self) -> dict:
        return self.get('pre_commits', {})

    def add_pre_commit(self, participant: str, pre_commit: PreCommitTransactionsMessage):
        if 'hash~sig' not in pre_commit:
            raise SiriusContextError(f'Pre-Commit for participant {participant} does not have hash~sig attribute')
        pre_commits = self.pre_commits
        pre_commits[participant] = pre_commit['hash~sig']
        self['pre_commits'] = pre_commits

    def validate(self):
        super().validate()
        for participant in self.participants:
            if participant not in self.pre_commits.keys():
                raise SiriusValidationError(f'Pre-Commit for participant "{participant}" does not exists')

    async def verify_pre_commits(self, api: AbstractCrypto, expected_state: MicroLedgerState):
        states = {}
        for participant, signed in self.pre_commits.items():
            state_hash, is_success = await verify_signed(api, signed)
            if not is_success:
                raise SiriusValidationError(f'Error verifying pre_commit for participant: {participant}')
            if state_hash != expected_state.hash:
                raise SiriusValidationError(f'Ledger state for participant {participant} is not consistent')
            states[participant] = (expected_state, signed)
        return states


class PostCommitTransactionsMessage(BaseTransactionsMessage):
    """Message to commit transactions list
    """
    NAME = 'stage-post-commit'

    @property
    def commits(self) -> List[dict]:
        payload = self.get('commits', [])
        if payload:
            return payload
        else:
            return []

    async def add_commit_sign(self, api: AbstractCrypto, commit: CommitTransactionsMessage, me: Pairwise.Me):
        signed = await sign(api, commit, me.verkey)
        commits = self.commits
        commits.append(signed)
        self['commits'] = commits

    async def verify_commits(self, api: AbstractCrypto, expected: CommitTransactionsMessage, verkeys: List[str]) -> bool:
        actual_verkeys = [commit['signer'] for commit in self.commits]
        if not set(verkeys).issubset(set(actual_verkeys)):
            return False
        for signed in self.commits:
            commit, is_success = await verify_signed(api, signed)
            if is_success:
                cleaned_commit = {k: v for k, v in commit.items() if not k.startswith('~')}
                cleaned_expect = {k: v for k, v in expected.items() if not k.startswith('~')}
                if cleaned_commit != cleaned_expect:
                    return False
            else:
                return False
        return True

    def validate(self):
        super().validate()
        if not self.commits:
            raise SiriusValidationError('Commits collection is empty')


class BaseParallelTransactionsMessage(SimpleConsensusMessage):

    NAME = 'stage-parallel'

    def __init__(
            self, transactions: List[Transaction] = None,
            states: List[Union[MicroLedgerState, dict]] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if transactions is not None:
            for txn in transactions:
                txn = Transaction(txn)
                if not txn.has_metadata():
                    raise SiriusContextError('Transaction must have metadata for specific Ledger')
            self['transactions'] = transactions
        if states:
            # Fix states as hash
            states = [MicroLedgerState(state) for state in states]
            # sort by ledger name, assume ledger name is unique in system
            # to make available to calc accumulated hash predictable
            states = list(sorted(states, key=lambda s: s.name))
            accum = hashlib.sha256()
            for state in states:
                accum.update(state.hash.encode())
            self['hash'] = accum.hexdigest()

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @property
    def transactions(self) -> Optional[List[Transaction]]:
        txns = self.get('transactions', None)
        if txns is not None:
            return [Transaction(txn) for txn in txns]
        else:
            return None

    @property
    def hash(self) -> Optional[str]:
        return self.get('hash', None)


class ProposeParallelTransactionsMessage(BaseParallelTransactionsMessage):
    """Message to process parallel transactions propose by Actor
    """
    NAME = 'stage-propose-parallel'

    def __init__(
            self, transactions: List[Transaction] = None, states: List[Union[MicroLedgerState, dict]] = None,
            timeout_sec: int = None, *args, **kwargs
    ):
        super().__init__(transactions=transactions, states=states, *args, **kwargs)
        if timeout_sec:
            self['timeout_sec'] = timeout_sec
        if states:
            states = [MicroLedgerState(state) for state in states]
            self['ledgers'] = [state.name for state in states]

    @property
    def ledgers(self) -> List[str]:
        return self.get('ledgers', [])

    @property
    def timeout_sec(self) -> Optional[int]:
        return self.get('timeout_sec', None)

    def validate(self):
        super().validate()
        if not self.transactions:
            raise SiriusValidationError('Empty transactions list')
        for txn in self.transactions:
            if not txn.has_metadata():
                raise SiriusValidationError('Transaction must have metadata')
        if not self.ledgers:
            raise SiriusValidationError('Ledgers is empty')
        if not self.hash:
            raise SiriusValidationError('Empty hash')


class PreCommitParallelTransactionsMessage(BaseParallelTransactionsMessage):
    """Message to accumulate participants signed accepts for transactions list in parallel mode
    """
    NAME = 'stage-pre-commit-parallel'

    async def sign_states(self, api: AbstractCrypto, me: Pairwise.Me):
        signed = await sign(api, self.hash, me.verkey)
        self['hash~sig'] = signed
        if 'transactions' in self:
            del self['transactions']

    async def verify_state(self, api: AbstractCrypto, expected_verkey: str) -> (bool, Optional[str]):
        hash_signed = self.get('hash~sig', None)
        if hash_signed:
            if hash_signed['signer'] == expected_verkey:
                state_hash, is_success = await verify_signed(api, hash_signed)
                return is_success, state_hash
            else:
                return False, None
        else:
            return False, None


class CommitParallelTransactionsMessage(BaseParallelTransactionsMessage):
    """Message to commit transactions list in parallel mode
    """
    NAME = 'stage-commit-parallel'

    @property
    def pre_commits(self) -> dict:
        return self.get('pre_commits', {})

    def add_pre_commit(self, participant: str, pre_commit: PreCommitParallelTransactionsMessage):
        if 'hash~sig' not in pre_commit:
            raise SiriusContextError(f'Pre-Commit for participant {participant} does not have hash~sig attribute')
        pre_commits = self.pre_commits
        pre_commits[participant] = pre_commit['hash~sig']
        self['pre_commits'] = pre_commits
        participants = self.participants
        participants.append(participant)
        self['participants'] = list(set(participants))

    def validate(self):
        super().validate()
        if not self.participants:
            raise SiriusValidationError('Participants list is empty')
        for participant in self.participants:
            if participant not in self.pre_commits.keys():
                raise SiriusValidationError(f'Pre-Commit for participant "{participant}" does not exists')

    async def verify_pre_commits(self, api: AbstractCrypto, expected_hash: str):
        states = {}
        for participant, signed in self.pre_commits.items():
            state_hash, is_success = await verify_signed(api, signed)
            if not is_success:
                raise SiriusValidationError(f'Error verifying pre_commit for participant: {participant}')
            if state_hash != expected_hash:
                raise SiriusValidationError(f'Ledger state for participant {participant} is not consistent')
            states[participant] = (expected_hash, signed)
        return states


class PostCommitParallelTransactionsMessage(BaseTransactionsMessage):
    """Message to commit transactions list in parallel mode
    """
    NAME = 'stage-post-commit-parallel'

    @property
    def commits(self) -> List[dict]:
        payload = self.get('commits', [])
        if payload:
            return payload
        else:
            return []

    async def add_commit_sign(self, api: AbstractCrypto, commit: CommitParallelTransactionsMessage, me: Pairwise.Me):
        signed = await sign(api, commit, me.verkey)
        commits = self.commits
        commits.append(signed)
        self['commits'] = commits

    async def verify_commits(self, api: AbstractCrypto, expected: CommitParallelTransactionsMessage, verkeys: List[str]) -> bool:
        actual_verkeys = [commit['signer'] for commit in self.commits]
        if not set(verkeys).issubset(set(actual_verkeys)):
            return False
        for signed in self.commits:
            commit, is_success = await verify_signed(api, signed)
            if is_success:
                cleaned_commit = {k: v for k, v in commit.items() if not k.startswith('~')}
                cleaned_expect = {k: v for k, v in expected.items() if not k.startswith('~')}
                if cleaned_commit != cleaned_expect:
                    return False
            else:
                return False
        return True

    def validate(self):
        super().validate()
        if not self.commits:
            raise SiriusValidationError('Commits collection is empty')


try:
    from collections.abc import MutableMapping
except ImportError:  # Will allow usage with virtually any python 3 version
    from collections import MutableMapping

from threading import Timer, Thread, Lock
from sortedcontainers import SortedKeyList
from time import time, sleep


class ExpiringDict(MutableMapping):
    def __init__(self, ttl=None, interval=0.100, *args, **kwargs):
        """
        Create an ExpiringDict class, optionally passing in a time-to-live
        number in seconds that will act globally as an expiration time for keys.

        If omitted, the dict will work like a normal dict by default, expiring
        only keys explicity set via the `.ttl` method.
        """
        self.__store = dict(*args, **kwargs)
        self.__keys = SortedKeyList(key=lambda x: x[0])
        self.__ttl = ttl
        self.__lock = Lock()
        self.__interval = interval

        Thread(target=self.__worker, daemon=True).start()

    def flush(self):
        now = time()
        max_index = 0
        with self.__lock:
            for index, (timestamp, key) in enumerate(self.__keys):
                if timestamp > now:  # rest of the timestamps in future
                    max_index = index
                    break
                try:
                    del self.__store[key]
                except KeyError:
                    pass  # don't care if it was deleted early
            del self.__keys[0:max_index]

    def __worker(self):
        while True:
            self.flush()
            sleep(self.__interval)

    def __setitem__(self, key, value):
        """
        Set `value` of `key` in dict. `key` will be automatically
        deleted if the `ttl` option was provided for this object.
        """
        if self.__ttl:
            self.__set_with_expire(key, value, self.__ttl)
        else:
            self.__store[key] = value

    def ttl(self, key, value, ttl):
        """
        Set `value` of `key` in dict to expire after `ttl` seconds.
        Overrides object level `ttl` setting.
        """
        self.__set_with_expire(key, value, ttl)

    def __set_with_expire(self, key, value, ttl):
        self.__lock.acquire()
        self.__keys.add((time() + ttl, key))
        self.__store[key] = value
        self.__lock.release()

    def __delitem__(self, key):
        del self.__store[key]

    def __getitem__(self, key):
        return self.__store[key]

    def __iter__(self):
        return iter(self.__store)

    def __len__(self):
        return len(self.__store)



from typing import Optional, Any, List

from sirius_sdk.storages.abstract.immutable import AbstractImmutableCollection


class InMemoryImmutableCollection(AbstractImmutableCollection):

    def __init__(self, *args, **kwargs):
        self.__databases = {}
        self.__selected_db = None
        super().__init__(*args, **kwargs)

    async def select_db(self, db_name: str):
        if db_name not in self.__databases:
            self.__databases[db_name] = []
        self.__selected_db = self.__databases[db_name]

    async def add(self, value: Any, tags: dict):
        item = (value, tags)
        self.__selected_db.append(item)

    async def fetch(self, tags: dict, limit: int = None) -> (List[Any], int):
        result = []
        for item in self.__selected_db:
            value_ = item[0]
            tags_ = item[1]
            if tags.items() <= tags_.items():
                result.append(value_)
        return result



import asyncio
from abc import ABC, abstractmethod
from typing import Optional, List, Any, Union, Tuple, Dict
from contextlib import asynccontextmanager

from sirius_sdk.agent.pairwise import Pairwise, TheirEndpoint
from sirius_sdk.agent.listener import Event
from sirius_sdk.messaging import Message
from sirius_sdk.errors.exceptions import SiriusContextError, OperationAbortedManually, SiriusConnectionClosed, \
    SiriusTimeoutIO

from .core import _current_hub


PLEASE_ACK_DECORATOR = '~please_ack'
THREAD_DECORATOR = '~thread'


class AbstractCoProtocol(ABC):

    def __init__(self, time_to_live: int = None):
        self.__time_to_live = time_to_live
        self.__is_aborted = False
        self._hub = None
        self._is_start = False
        self._transport = None

    def __del__(self):
        if self._is_start and self._hub:
            self._hub.run_soon(self._transport.stop())

    @property
    def time_to_live(self) -> Optional[int]:
        return self.__time_to_live

    @property
    def is_aborted(self) -> bool:
        return self.__is_aborted

    async def abort(self):
        if self._hub:
            self._hub.run_soon(self.__clean())
            if not self.__is_aborted:
                self.__is_aborted = True
                await self._hub.abort()
                self._hub = None

    async def __clean(self):
        if self._is_start:
            await self._transport.stop()
            self._is_start = False
        self._transport = None


class AbstractP2PCoProtocol(AbstractCoProtocol):

    def __init__(self, time_to_live: int = None):
        super().__init__(time_to_live)

    @abstractmethod
    async def send(self, message: Message):
        pass

    @abstractmethod
    async def get_one(self) -> (Optional[Message], str, Optional[str]):
        """Accumulate event from participant

        return: message, sender_verkey, recipient_verkey
        """
        pass

    @abstractmethod
    async def switch(self, message: Message) -> (bool, Message):
        pass


class CoProtocolP2PAnon(AbstractP2PCoProtocol):

    def __init__(self, my_verkey: str, endpoint: TheirEndpoint, protocols: List[str], time_to_live: int = None):
        if not protocols:
            raise SiriusContextError('You must set protocols list. It is empty for now!')
        super().__init__(time_to_live=time_to_live)
        self.__my_verkey = my_verkey
        self.__endpoint = endpoint
        self.__thread_id = None
        self.__protocols = protocols

    @property
    def protocols(self) -> List[str]:
        return self.__protocols

    async def send(self, message: Message):
        async with self.__get_transport_lazy() as transport:
            self.__setup(message, please_ack=False)
            await transport.send(message)

    async def get_one(self) -> (Optional[Message], str, Optional[str]):
        """
        return message, sender_verkey, recipient_verkey
        """
        async with self.__get_transport_lazy() as transport:
            message, sender_verkey, recipient_verkey = await transport.get_one()
        return message, sender_verkey, recipient_verkey

    async def switch(self, message: Message) -> (bool, Message):
        async with self.__get_transport_lazy() as transport:
            self.__setup(message)
            success, response = await transport.switch(message)
            if success:
                if PLEASE_ACK_DECORATOR in response:
                    self.__thread_id = response.get(PLEASE_ACK_DECORATOR, {}).get('message_id', None) or message.id
                else:
                    self.__thread_id = None
        return success, response

    def __setup(self, message: Message, please_ack: bool = True):
        if please_ack:
            if PLEASE_ACK_DECORATOR not in message:
                message[PLEASE_ACK_DECORATOR] = {'message_id': message.id}
        if self.__thread_id:
            thread = message.get(THREAD_DECORATOR, {})
            if 'thid' not in thread:
                thread['thid'] = self.__thread_id
                message[THREAD_DECORATOR] = thread

    @asynccontextmanager
    async def __get_transport_lazy(self):
        if self._transport is None:
            self._hub = _current_hub()
            async with self._hub.get_agent_connection_lazy() as agent:
                self._transport = await agent.spawn(self.__my_verkey, self.__endpoint)
                await self._transport.start(protocols=self.protocols, time_to_live=self.time_to_live)
                self._is_start = True
        try:
            yield self._transport
        except SiriusConnectionClosed:
            if self.is_aborted:
                raise OperationAbortedManually('User aborted operation')
            else:
                raise


class CoProtocolP2P(AbstractP2PCoProtocol):

    def __init__(self, pairwise: Pairwise, protocols: List[str], time_to_live: int = None):
        if not protocols:
            raise SiriusContextError('You must set protocols list. It is empty for now!')
        super().__init__(time_to_live=time_to_live)
        self.__pairwise = pairwise
        self.__thread_id = None
        self.__protocols = protocols

    @property
    def protocols(self) -> List[str]:
        return self.__protocols

    async def send(self, message: Message):
        async with self.__get_transport_lazy() as transport:
            self.__setup(message, please_ack=False)
            await transport.send(message)

    async def get_one(self) -> (Optional[Message], str, Optional[str]):
        """
        return message, sender_verkey, recipient_verkey
        """
        async with self.__get_transport_lazy() as transport:
            message, sender_verkey, recipient_verkey = await transport.get_one()
        return message, sender_verkey, recipient_verkey

    async def switch(self, message: Message) -> (bool, Message):
        async with self.__get_transport_lazy() as transport:
            self.__setup(message)
            success, response = await transport.switch(message)
            if PLEASE_ACK_DECORATOR in response:
                self.__thread_id = response.get(PLEASE_ACK_DECORATOR, {}).get('message_id', None) or message.id
            else:
                self.__thread_id = None
        return success, response

    def __setup(self, message: Message, please_ack: bool = True):
        if please_ack:
            if PLEASE_ACK_DECORATOR not in message:
                message[PLEASE_ACK_DECORATOR] = {'message_id': message.id}
        if self.__thread_id:
            thread = message.get(THREAD_DECORATOR, {})
            if 'thid' not in thread:
                thread['thid'] = self.__thread_id
                message[THREAD_DECORATOR] = thread

    @asynccontextmanager
    async def __get_transport_lazy(self):
        if self._transport is None:
            self._hub = _current_hub()
            async with self._hub.get_agent_connection_lazy() as agent:
                self._transport = await agent.spawn(self.__pairwise)
                await self._transport.start(protocols=self.protocols, time_to_live=self.time_to_live)
                self._is_start = True
        try:
            yield self._transport
        except SiriusConnectionClosed:
            if self.is_aborted:
                raise OperationAbortedManually('User aborted operation')
            else:
                raise


class CoProtocolThreadedP2P(AbstractP2PCoProtocol):

    def __init__(self, thid: str, to: Pairwise, pthid: str = None, time_to_live: int = None):
        super().__init__(time_to_live=time_to_live)
        self.__thid = thid
        self.__pthid = pthid
        self.__to = to

    @property
    def thid(self) -> str:
        return self.__thid

    @property
    def pthid(self) -> Optional[str]:
        return self.__pthid

    async def send(self, message: Message):
        async with self.__get_transport_lazy() as transport:
            await transport.send(message)

    async def get_one(self) -> (Optional[Message], str, Optional[str]):
        """
        return message, sender_verkey, recipient_verkey
        """
        async with self.__get_transport_lazy() as transport:
            message, sender_verkey, recipient_verkey = await transport.get_one()
        return message, sender_verkey, recipient_verkey

    async def switch(self, message: Message) -> (bool, Message):
        async with self.__get_transport_lazy() as transport:
            success, response = await transport.switch(message)
            return success, response

    @asynccontextmanager
    async def __get_transport_lazy(self):
        if self._transport is None:
            self._hub = _current_hub()
            async with self._hub.get_agent_connection_lazy() as agent:
                if self.__pthid is None:
                    self._transport = await agent.spawn(self.__thid, self.__to)
                else:
                    self._transport = await agent.spawn(self.__thid, self.__to, self.__pthid)
                await self._transport.start(time_to_live=self.time_to_live)
                self._is_start = True
        try:
            yield self._transport
        except SiriusConnectionClosed:
            if self.is_aborted:
                raise OperationAbortedManually('User aborted operation')
            else:
                raise


class CoProtocolThreadedTheirs(AbstractCoProtocol):

    def __init__(self, thid: str, theirs: List[Pairwise], pthid: str = None, time_to_live: int = None):
        if len(theirs) < 1:
            raise SiriusContextError('theirs is empty')
        super().__init__(time_to_live=time_to_live)
        self.__thid = thid
        self.__pthid = pthid
        self.__theirs = theirs
        self.__dids = [their.their.did for their in theirs]

    @property
    def theirs(self) -> List[Pairwise]:
        return self.__theirs

    async def send(self, message: Message) -> Dict[Pairwise, Tuple[bool, str]]:
        """Send message to given participants

        return: List[( str: participant-id, bool: message was successfully sent, str: endpoint response body )]
        """
        results = {}
        async with self.__get_transport_lazy() as transport:
            responses = await transport.send_many(message, self.__theirs)
        for p2p, response in zip(self.__theirs, responses):
            success, body = response
            results[p2p] = (success, body)
        return results

    async def get_one(self) -> Tuple[Optional[Pairwise], Optional[Message]]:
        """Read event from any of participants at given timeout

        return: (Pairwise: participant-id, Message: message from given participant)
        """
        async with self.__get_transport_lazy() as transport:
            try:
                message, sender_verkey, recipient_verkey = await transport.get_one()
            except SiriusTimeoutIO:
                return None, None
            else:
                p2p = self.__load_p2p_from_verkey(sender_verkey)
                return p2p, message

    async def switch(self, message: Message) -> Dict[Pairwise, Tuple[bool, Optional[Message]]]:
        """Switch state while participants at given timeout give responses

        return: {
            Pairwise: participant,
            (
              bool: message was successfully sent to participant,
              Message: response message from participant or Null if request message was not successfully sent
            )
        }
        """
        statuses = await self.send(message)
        # fill errors to result just now
        results = {p2p: (False, None) for p2p, stat in statuses.items() if stat[0] is True}
        # then work with success participants only
        success_theirs = {p2p: (False, None) for p2p, stat in statuses.items() if stat[0] is True}
        accum = 0
        while accum < len(success_theirs):
            p2p, message = await self.get_one()
            if p2p is None:
                break
            if p2p and p2p.their.did in self.__dids:
                success_theirs[p2p] = (True, message)
                accum += 1
        results.update(success_theirs)
        return results

    def __load_p2p_from_verkey(self, verkey: str) -> Optional[Pairwise]:
        for p2p in self.__theirs:
            if p2p.their.verkey == verkey:
                return p2p
        return None

    @asynccontextmanager
    async def __get_transport_lazy(self):
        if self._transport is None:
            async with _current_hub().get_agent_connection_lazy() as agent:
                if self.__pthid is None:
                    self._transport = await agent.spawn(self.__thid)
                else:
                    self._transport = await agent.spawn(self.__thid, self.__pthid)
                await self._transport.start(time_to_live=self.time_to_live)
                self._is_start = True
        try:
            yield self._transport
        except SiriusConnectionClosed:
            if self.is_aborted:
                raise OperationAbortedManually('User aborted operation')
            else:
                raise


async def open_communication(event: Event, time_to_live: int = None) -> Optional[AbstractP2PCoProtocol]:
    if event.pairwise is not None and event.message is not None:
        thread_id = None
        parent_thread_id = None
        if THREAD_DECORATOR in event.message:
            thread_id = event.message.get(THREAD_DECORATOR, {}).get('thid', None)
        if PLEASE_ACK_DECORATOR in event.message:
            parent_thread_id = thread_id
            thread_id = event.message.get(PLEASE_ACK_DECORATOR, {}).get('message_id', None) or event.message.id
        if thread_id:
            comm = CoProtocolThreadedP2P(
                thid=thread_id,
                to=event.pairwise,
                pthid=parent_thread_id,
                time_to_live=time_to_live
            )
        else:
            comm = CoProtocolP2P(
                pairwise=event.pairwise,
                protocols=[event.message.protocol],
                time_to_live=time_to_live
            )
        return comm
    else:
        return None



import json
from urllib.parse import urlparse, urlunparse

import pytest

import sirius_sdk
from sirius_sdk.agent.connections import Endpoint
from sirius_sdk.agent.aries_rfc.feature_0160_connection_protocol.state_machines import Inviter, Invitee, \
    ConnRequest, Invitation

from .helpers import run_coroutines, IndyAgent, ServerTestSuite


def replace_url_components(url: str, base: str = None) -> str:
    ret = url
    if base:
        parsed = urlparse(url)
        components = list(parsed)
        components[1] = urlparse(base).netloc
        ret = urlunparse(components)
    return ret


async def read_events(uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection):
    async with sirius_sdk.context(uri, credentials, p2p):
        listener = await sirius_sdk.subscribe()
        async for event in listener:
            print('========= EVENT ============')
            print(json.dumps(event, indent=2, sort_keys=True))
            print('============================')


async def run_inviter(
        uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, expected_connection_key: str,
        me: sirius_sdk.Pairwise.Me = None, replace_endpoints: bool = False, did_doc_extra: dict = None
):
    async with sirius_sdk.context(uri, credentials, p2p):
        endpoints_ = await sirius_sdk.endpoints()
        my_endpoint = [e for e in endpoints_ if e.routing_keys == []][0]
        if replace_endpoints:
            new_address = replace_url_components(my_endpoint.address, pytest.test_suite_overlay_address)
            my_endpoint = Endpoint(new_address, my_endpoint.routing_keys, is_default=my_endpoint.is_default)
        listener = await sirius_sdk.subscribe()
        async for event in listener:
            connection_key = event['recipient_verkey']
            if expected_connection_key == connection_key:
                request = event['message']
                assert isinstance(request, ConnRequest)
                if replace_endpoints:
                    request['connection']['did_doc']['service'][0]['serviceEndpoint'] = replace_url_components(
                        request['connection']['did_doc']['service'][0]['serviceEndpoint'],
                        pytest.old_agent_overlay_address
                    )
                # Setup state machine
                if me is None:
                    my_did, my_verkey = await sirius_sdk.DID.create_and_store_my_did()
                    me = sirius_sdk.Pairwise.Me(did=my_did, verkey=my_verkey)
                # create connection
                machine = Inviter(me, connection_key, my_endpoint)
                ok, pairwise = await machine.create_connection(request, did_doc_extra)
                assert ok is True
                await sirius_sdk.PairwiseList.ensure_exists(pairwise)
        pass


async def run_invitee(
        uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        invitation: Invitation, my_label: str, me: sirius_sdk.Pairwise.Me = None,
        replace_endpoints: bool = False, did_doc_extra: dict = None
):
    async with sirius_sdk.context(uri, credentials, p2p):
        if me is None:
            my_did, my_verkey = await sirius_sdk.DID.create_and_store_my_did()
            me = sirius_sdk.Pairwise.Me(did=my_did, verkey=my_verkey)
        endpoints_ = await sirius_sdk.endpoints()
        my_endpoint = [e for e in endpoints_ if e.routing_keys == []][0]
        if replace_endpoints:
            new_address = replace_url_components(my_endpoint.address, pytest.test_suite_overlay_address)
            my_endpoint = Endpoint(new_address, my_endpoint.routing_keys, is_default=my_endpoint.is_default)
            new_address = replace_url_components(invitation['serviceEndpoint'], pytest.old_agent_overlay_address)
            invitation['serviceEndpoint'] = new_address
        # Create and start machine
        machine = Invitee(me, my_endpoint)
        ok, pairwise = await machine.create_connection(invitation=invitation, my_label=my_label, did_doc=did_doc_extra)
        assert ok is True
        await sirius_sdk.PairwiseList.ensure_exists(pairwise)


@pytest.mark.asyncio
async def test_establish_connection(test_suite: ServerTestSuite):
    inviter = test_suite.get_agent_params('agent1')
    invitee = test_suite.get_agent_params('agent2')

    # Get endpoints
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        inviter_endpoint_address = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        connection_key = await sirius_sdk.Crypto.create_key()
        invitation = Invitation(label='Inviter', endpoint=inviter_endpoint_address, recipient_keys=[connection_key])

    # Init Me
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        inviter_me = sirius_sdk.Pairwise.Me(did, verkey)
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        invitee_me = sirius_sdk.Pairwise.Me(did, verkey)

    await run_coroutines(
        run_inviter(
            inviter['server_address'], inviter['credentials'], inviter['p2p'], connection_key, inviter_me
        ),
        run_invitee(
            invitee['server_address'], invitee['credentials'], invitee['p2p'], invitation, 'Invitee', invitee_me
        )
    )

    # Check for Inviter
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        pairwise = await sirius_sdk.PairwiseList.load_for_verkey(invitee_me.verkey)
        assert pairwise is not None
        assert pairwise.their.did == invitee_me.did
    # Check for Invitee
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        pairwise = await sirius_sdk.PairwiseList.load_for_verkey(inviter_me.verkey)
        assert pairwise is not None
        assert pairwise.their.did == inviter_me.did


@pytest.mark.asyncio
async def test_update_pairwise_metadata(test_suite: ServerTestSuite):
    inviter = test_suite.get_agent_params('agent1')
    invitee = test_suite.get_agent_params('agent2')

    # Get endpoints
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        inviter_endpoint_address = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        connection_key = await sirius_sdk.Crypto.create_key()
        invitation = Invitation(label='Inviter', endpoint=inviter_endpoint_address, recipient_keys=[connection_key])
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        invitee_endpoint_address = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address

    # Init Me
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        inviter_side = sirius_sdk.Pairwise.Me(did, verkey)
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        invitee_side = sirius_sdk.Pairwise.Me(did, verkey)
    # Manually set pairwise list
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        await sirius_sdk.DID.store_their_did(invitee_side.did, invitee_side.verkey)
        p = sirius_sdk.Pairwise(
            me=inviter_side,
            their=sirius_sdk.Pairwise.Their(
                invitee_side.did, 'Invitee', invitee_endpoint_address, invitee_side.verkey
            )
        )
        await sirius_sdk.PairwiseList.create(p)
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        await sirius_sdk.DID.store_their_did(inviter_side.did, inviter_side.verkey)
        p = sirius_sdk.Pairwise(
            me=invitee_side,
            their=sirius_sdk.Pairwise.Their(
                inviter_side.did, 'Inviter', inviter_endpoint_address, inviter_side.verkey
            )
        )
        await sirius_sdk.PairwiseList.create(p)

    await run_coroutines(
        run_inviter(
            inviter['server_address'], inviter['credentials'], inviter['p2p'], connection_key, inviter_side
        ),
        run_invitee(
            invitee['server_address'], invitee['credentials'], invitee['p2p'], invitation, 'Invitee', invitee_side
        )
    )

    # Check for Inviter
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        pairwise = await sirius_sdk.PairwiseList.load_for_did(invitee_side.did)
        assert pairwise.metadata != {}
        assert pairwise.metadata is not None

    # Check for Invitee
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        pairwise = await sirius_sdk.PairwiseList.load_for_did(inviter_side.did)
        assert pairwise.metadata != {}
        assert pairwise.metadata is not None


@pytest.mark.skip
@pytest.mark.asyncio
async def test_invitee_back_compatibility(indy_agent: IndyAgent, test_suite: ServerTestSuite):
    their_invitaton = await indy_agent.create_invitation(label='Test Invitee')
    invitation = Invitation.from_url(their_invitaton['url'])
    invitee = test_suite.get_agent_params('agent1')

    # Init invitee
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        invitee_side = sirius_sdk.Pairwise.Me(did, verkey)

    await run_coroutines(
        run_invitee(
            invitee['server_address'], invitee['credentials'], invitee['p2p'], invitation, 'Invitee', invitee_side, True
        ),
        read_events(
            invitee['server_address'], invitee['credentials'], invitee['p2p']
        )
    )
    invitation_pairwise = None
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        async for i, pairwise in sirius_sdk.PairwiseList.enumerate():
            if pairwise.me.did == invitee_side.did:
                invitation_pairwise = pairwise
                break
    assert invitation_pairwise is not None


@pytest.mark.skip
@pytest.mark.asyncio
async def test_inviter_back_compatibility(indy_agent: IndyAgent, test_suite: ServerTestSuite, agent1: sirius_sdk.Agent):
    inviter = test_suite.get_agent_params('agent1')
    # Init inviter
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        inviter_endpoint_address = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        connection_key = await sirius_sdk.Crypto.create_key()
        inviter_endpoint_address = replace_url_components(inviter_endpoint_address, pytest.test_suite_overlay_address)
        invitation = Invitation(label='Inviter', endpoint=inviter_endpoint_address, recipient_keys=[connection_key])
        invitation_url = invitation.invitation_url
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        inviter_side = sirius_sdk.Pairwise.Me(did, verkey)

    await run_coroutines(
        run_inviter(
            inviter['server_address'], inviter['credentials'], inviter['p2p'], connection_key, inviter_side, True
        ),
        indy_agent.invite(invitation_url=invitation_url),
    )

    invitated_pairwise = None
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        async for i, p in sirius_sdk.PairwiseList.enumerate():
            assert isinstance(p, sirius_sdk.Pairwise)
            if p.me.did == inviter_side.did:
                invitated_pairwise = p
                break
    assert invitated_pairwise is not None


@pytest.mark.asyncio
async def test_did_doc_extra_fields(test_suite: ServerTestSuite):
    inviter = test_suite.get_agent_params('agent1')
    invitee = test_suite.get_agent_params('agent2')

    # Get endpoints
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        inviter_endpoint_address = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        connection_key = await sirius_sdk.Crypto.create_key()
        invitation = Invitation(label='Inviter', endpoint=inviter_endpoint_address, recipient_keys=[connection_key])

    # Init Me
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        inviter_me = sirius_sdk.Pairwise.Me(did, verkey)
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        did, verkey = await sirius_sdk.DID.create_and_store_my_did()
        invitee_me = sirius_sdk.Pairwise.Me(did, verkey)

    await run_coroutines(
        run_inviter(
            inviter['server_address'], inviter['credentials'], inviter['p2p'], connection_key, inviter_me,
            did_doc_extra={
                'creator': {'@id': 'uuid-xxx-yyy'},
                'extra': 'Any'
            }
        ),
        run_invitee(
            invitee['server_address'], invitee['credentials'], invitee['p2p'], invitation, 'Invitee', invitee_me,
            did_doc_extra={
                'creator': {'@id': 'uuid-www-zzz'},
                'extra': 'Test'
            }
        )
    )

    # Check for Inviter
    async with sirius_sdk.context(inviter['server_address'], inviter['credentials'], inviter['p2p']):
        pairwise = await sirius_sdk.PairwiseList.load_for_verkey(invitee_me.verkey)
        assert pairwise is not None
        assert pairwise.their.did == invitee_me.did
        assert pairwise.me.did_doc is not None
        assert pairwise.me.did_doc.get('creator', {}) == {'@id': 'uuid-xxx-yyy'}
        assert pairwise.me.did_doc.get('extra', None) == 'Any'
        assert pairwise.their.did_doc is not None
        assert pairwise.their.did_doc.get('creator', {}) == {'@id': 'uuid-www-zzz'}
        assert pairwise.their.did_doc.get('extra', None) == 'Test'
    # Check for Invitee
    async with sirius_sdk.context(invitee['server_address'], invitee['credentials'], invitee['p2p']):
        pairwise = await sirius_sdk.PairwiseList.load_for_verkey(inviter_me.verkey)
        assert pairwise is not None
        assert pairwise.their.did == inviter_me.did
        assert pairwise.me.did_doc is not None
        assert pairwise.me.did_doc.get('creator', {}) == {'@id': 'uuid-www-zzz'}
        assert pairwise.me.did_doc.get('extra', None) == 'Test'
        assert pairwise.their.did_doc is not None
        assert pairwise.their.did_doc.get('creator', {}) == {'@id': 'uuid-xxx-yyy'}
        assert pairwise.their.did_doc.get('extra', None) == 'Any'



import aiohttp


async def http_send(
        msg: bytes, endpoint: str, timeout: float,
        connector: aiohttp.TCPConnector = None,
        content_type: str = 'application/ssi-agent-wire'
):
    """Send over HTTP"""
    headers = {'content-type': content_type}
    request = aiohttp.request('post', endpoint, data=msg, headers=headers, connector=connector)
    async with request as resp:
        body = await resp.read()
        if resp.status in [200, 202]:
            return True, body
        else:
            return False, body



from enum import Enum

from sirius_sdk.agent.wallet.abstract.did import *
from sirius_sdk.agent.wallet.abstract.cache import *
from sirius_sdk.agent.wallet.abstract.crypto import *
from sirius_sdk.agent.wallet.abstract.ledger import *
from sirius_sdk.agent.wallet.abstract.pairwise import *
from sirius_sdk.agent.wallet.abstract.anoncreds import *
from sirius_sdk.agent.wallet.abstract.non_secrets import *


class KeyDerivationMethod(Enum):

    DEFAULT = 'ARGON2I_MOD'
    FAST = 'ARGON2I_INT'
    RAW = 'RAW'

    def serialize(self):
        return self.value

    @staticmethod
    def deserialize(buffer: str):
        value = buffer
        if value == 'ARGON2I_MOD':
            return KeyDerivationMethod.DEFAULT
        elif value == 'ARGON2I_INT':
            return KeyDerivationMethod.FAST
        elif value == 'RAW':
            return KeyDerivationMethod.RAW
        else:
            raise RuntimeError('Unexpected value "%s"' % buffer)


__all__ = ["KeyDerivationMethod"]



from typing import Optional

from sirius_sdk.agent.aries_rfc.base import THREAD_DECORATOR


class PleaseAckMixin:

    @property
    def ack_message_id(self) -> str:
        return self.get('~please_ack', {}).get('message_id', None) or self.id

    @property
    def please_ack(self) -> bool:
        """https://github.com/hyperledger/aries-rfcs/tree/master/features/0317-please-ack"""
        return self.get('~please_ack', None) is not None

    @please_ack.setter
    def please_ack(self, flag: bool):
        if flag:
            self['~please_ack'] = {'message_id': self.id}
        elif '~please_ack' in self:
            del self['~please_ack']


class ThreadMixin:

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @thread_id.setter
    def thread_id(self, thid: str):
        thread = self.get(THREAD_DECORATOR, {})
        thread['thid'] = thid
        self[THREAD_DECORATOR] = thread



from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.state_machines import Verifier, Prover
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.messages import ProposedAttrib, ProposedPredicate, \
    AttribTranslation, PresentProofProblemReport, RequestPresentationMessage, PresentationMessage


__all__ = [
    'Verifier', 'Prover', 'ProposedAttrib', 'ProposedPredicate', 'AttribTranslation', 'PresentProofProblemReport',
    'RequestPresentationMessage', 'PresentationMessage'
]



from sirius_sdk.agent.consensus.simple.messages import InitRequestLedgerMessage, InitResponseLedgerMessage, MicroLedgerState, \
    ProposeTransactionsMessage, PreCommitTransactionsMessage, CommitTransactionsMessage, PostCommitTransactionsMessage
from sirius_sdk.agent.consensus.simple.state_machines import MicroLedgerSimpleConsensus


__all__ = [
    'MicroLedgerSimpleConsensus', 'InitRequestLedgerMessage',
    'InitResponseLedgerMessage', 'MicroLedgerState',
    'ProposeTransactionsMessage', 'PreCommitTransactionsMessage',
    'CommitTransactionsMessage', 'PostCommitTransactionsMessage'
]



import json
from abc import ABC, abstractmethod
from typing import Optional, List

from sirius_sdk.base import JsonSerializable


class RetrieveRecordOptions(JsonSerializable):

    def __init__(self, retrieve_type: bool=False, retrieve_value: bool=False, retrieve_tags: bool=False):
        self.retrieve_type = retrieve_type
        self.retrieve_value = retrieve_value
        self.retrieve_tags = retrieve_tags

    def check_all(self):
        self.retrieve_type = True
        self.retrieve_value = True
        self.retrieve_tags = True

    def to_json(self):
        options = dict()
        if self.retrieve_type:
            options['retrieveType'] = self.retrieve_type
        if self.retrieve_value:
            options['retrieveValue'] = self.retrieve_value
        if self.retrieve_tags:
            options['retrieveTags'] = self.retrieve_tags
        return options

    def serialize(self):
        return json.dumps(self.to_json())

    def deserialize(self, buffer: str):
        data = json.loads(buffer)
        self.retrieve_type = data.get('retrieveType', False)
        self.retrieve_value = data.get('retrieveValue', False)
        self.retrieve_tags = data.get('retrieveTags', False)


class AbstractNonSecrets(ABC):

    @abstractmethod
    async def add_wallet_record(self, type_: str, id_: str, value: str, tags: dict=None) -> None:
        """
        Create a new non-secret record in the wallet

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :param value: the value of record
        :param tags: the record tags used for search and storing meta information as json:
           {
             "tagName1": <str>, // string tag (will be stored encrypted)
             "tagName2": <str>, // string tag (will be stored encrypted)
             "~tagName3": <str>, // string tag (will be stored un-encrypted)
             "~tagName4": <str>, // string tag (will be stored un-encrypted)
           }
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def update_wallet_record_value(self, type_: str, id_: str, value: str) -> None:
        """
        Update a non-secret wallet record value

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :param value: the value of record
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def update_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        """
        Update a non-secret wallet record value

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :param tags: ags_json: the record tags used for search and storing meta information as json:
           {
             "tagName1": <str>, // string tag (will be stored encrypted)
             "tagName2": <str>, // string tag (will be stored encrypted)
             "~tagName3": <str>, // string tag (will be stored un-encrypted)
             "~tagName4": <str>, // string tag (will be stored un-encrypted)
           }
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def add_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        """
        Add new tags to the wallet record

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :param tags: ags_json: the record tags used for search and storing meta information as json:
           {
             "tagName1": <str>, // string tag (will be stored encrypted)
             "tagName2": <str>, // string tag (will be stored encrypted)
             "~tagName3": <str>, // string tag (will be stored un-encrypted)
             "~tagName4": <str>, // string tag (will be stored un-encrypted)
           }
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def delete_wallet_record_tags(self, type_: str, id_: str, tag_names: List[str]) -> None:
        """
        Add new tags to the wallet record

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :param tag_names: the list of tag names to remove from the record as json array: ["tagName1", "tagName2", ...]
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def delete_wallet_record(self, type_: str, id_: str) -> None:
        """
        Delete an existing wallet record in the wallet

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :return: None
        """
        raise NotImplemented

    @abstractmethod
    async def get_wallet_record(self, type_: str, id_: str, options: RetrieveRecordOptions) -> Optional[dict]:
        """
        Get an wallet record by id

        :param type_: allows to separate different record types collections
        :param id_: the id of record
        :param options:
          {
            retrieveType: (optional, false by default) Retrieve record type,
            retrieveValue: (optional, true by default) Retrieve record value,
            retrieveTags: (optional, true by default) Retrieve record tags
          }
        :return: wallet record json:
         {
           id: "Some id",
           type: "Some type", // present only if retrieveType set to true
           value: "Some value", // present only if retrieveValue set to true
           tags: <tags json>, // present only if retrieveTags set to true
         }
        """
        raise NotImplemented

    @abstractmethod
    async def wallet_search(self, type_: str, query: dict, options: RetrieveRecordOptions, limit: int=1) -> (List[dict],int):
        """
        Search for wallet records

        :param type_: allows to separate different record types collections
        :param query: MongoDB style query to wallet record tags:
          {
            "tagName": "tagValue",
            $or: {
              "tagName2": { $regex: 'pattern' },
              "tagName3": { $gte: '123' },
            },
          }
        :param options:
          {
            retrieveRecords: (optional, true by default) If false only "counts" will be calculated,
            retrieveTotalCount: (optional, false by default) Calculate total count,
            retrieveType: (optional, false by default) Retrieve record type,
            retrieveValue: (optional, true by default) Retrieve record value,
            retrieveTags: (optional, true by default) Retrieve record tags,
          }
        :param limit: max record count to retrieve
        :return: wallet records json:
         {
           totalCount: <str>, // present only if retrieveTotalCount set to true
           records: [{ // present only if retrieveRecords set to true
               id: "Some id",
               type: "Some type", // present only if retrieveType set to true
               value: "Some value", // present only if retrieveValue set to true
               tags: <tags json>, // present only if retrieveTags set to true
           }],
         }
        """
        raise NotImplemented



r"""It is package that implements features and concepts of Aries RFCs

see Aries RFC here: https://github.com/hyperledger/aries-rfcs
"""

from .feature_0015_acks import *
from .feature_0036_issue_credential import *
from .feature_0037_present_proof import *
from .feature_0048_trust_ping import *
from .feature_0095_basic_message import *
from .feature_0113_question_answer import *
from .feature_0160_connection_protocol import *



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Union, List
from urllib.parse import urljoin
from inspect import iscoroutinefunction

import aiohttp

from sirius_sdk.messaging import Message
from sirius_sdk.errors.exceptions import *


class JsonSerializable:

    @abstractmethod
    def serialize(self) -> dict:
        raise NotImplemented

    @classmethod
    @abstractmethod
    def deserialize(cls, buffer: Union[dict, bytes, str]):
        raise NotImplemented


class ReadOnlyChannel(ABC):
    """Communication abstraction for reading data stream
    """

    @abstractmethod
    async def read(self, timeout: int=None) -> bytes:
        """Read message packet

        :param timeout: Operation timeout is sec
        :return: chunk of data stream
        """
        raise NotImplemented()


class WriteOnlyChannel(ABC):
    """Communication abstraction for writing data stream
    """

    @abstractmethod
    async def write(self, data: bytes) -> bool:
        """
        Write message packet

        :param data: message packet
        :return: True if success ele False
        """
        raise NotImplemented()


class BaseConnector(ReadOnlyChannel, WriteOnlyChannel):
    """Transport Layer.

    Connectors operate as transport provider for high-level abstractions
    """

    @abstractmethod
    async def open(self):
        """Open communication
        """
        raise NotImplemented()

    @abstractmethod
    async def close(self):
        """Close communication
        """
        raise NotImplemented()


class WebSocketConnector(BaseConnector):

    DEF_TIMEOUT = 30.0
    ENC = 'utf-8'

    def __init__(
            self, server_address: str, path: str, credentials: bytes,
            timeout: float=DEF_TIMEOUT, loop: asyncio.AbstractEventLoop=None
    ):
        self.__session = aiohttp.ClientSession(
            loop=loop,
            timeout=aiohttp.ClientTimeout(total=timeout),
            headers={
                'origin': server_address,
                'credentials': credentials.decode('ascii')
            }
        )
        self._url = urljoin(server_address, path)
        self._ws = None

    def __del__(self):
        asyncio.ensure_future(self.__session.close())

    @property
    def is_open(self):
        return self._ws is not None and not self._ws.closed

    async def open(self):
        if not self.is_open:
            self._ws = await self.__session.ws_connect(url=self._url)

    async def close(self):
        if self.is_open:
            await self._ws.close()
            self._ws = None

    async def reopen(self):
        await self.close()
        await self.open()

    async def read(self, timeout: int=None) -> bytes:
        try:
            msg = await self._ws.receive(timeout=timeout)
        except asyncio.TimeoutError as e:
            raise SiriusTimeoutIO() from e
        if msg.type in [aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSED]:
            raise SiriusConnectionClosed()
        elif msg.type == aiohttp.WSMsgType.TEXT:
            return msg.data.encode(self.ENC)
        elif msg.type == aiohttp.WSMsgType.BINARY:
            return msg.data
        elif msg.type == aiohttp.WSMsgType.ERROR:
            raise SiriusIOError()

    async def write(self, message: Union[Message, bytes]) -> bool:
        if isinstance(message, Message):
            payload = message.serialize().encode(self.ENC)
        else:
            payload = message
        await self._ws.send_bytes(payload)
        return True


class AbstractStateMachine(ABC):

    def __init__(self, time_to_live: int = 60, logger=None, *args, **kwargs):
        """
        :param time_to_live: state machine time to live to finish progress
        """
        self.__time_to_live = time_to_live
        self.__is_aborted = False
        if logger is not None:
            if iscoroutinefunction(logger) or callable(logger):
                pass
            else:
                raise RuntimeError('Expect logger is iscoroutine function or callable object')
        self.__logger = logger
        self.__coprotocols = []

    @property
    def time_to_live(self) -> int:
        return self.__time_to_live

    @property
    def is_aborted(self) -> bool:
        return self.__is_aborted

    async def abort(self):
        """Abort state-machine"""
        self.__is_aborted = True
        for co in self.__coprotocols:
            await co.abort()
        self.__coprotocols.clear()

    async def log(self, **kwargs) -> bool:
        if self.__logger:
            kwargs = dict(**kwargs)
            kwargs['state_machine_id'] = id(self)
            await self.__logger(**kwargs)
        else:
            return False

    def _register_for_aborting(self, co):
        self.__coprotocols.append(co)

    def _unregister_for_aborting(self, co):
        self.__coprotocols = [item for item in self.__coprotocols if id(item) != id(co)]



import uuid
from typing import List, Any
from datetime import datetime, timedelta

import pytest

import sirius_sdk
from sirius_sdk.agent.wallet import NYMRole
from sirius_sdk.agent.aries_rfc.utils import str_to_utc
from sirius_sdk.agent.ledger import Schema, CredentialDefinition
from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential.state_machines import Issuer, Holder, \
    AttribTranslation, ProposedAttrib, OfferCredentialMessage
from sirius_sdk.errors.indy_exceptions import AnoncredsMasterSecretDuplicateNameError

from .conftest import get_pairwise
from .helpers import run_coroutines, IndyAgent, ServerTestSuite


async def run_issuer(
        uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        holder: sirius_sdk.Pairwise, values: dict, schema: Schema, cred_def: CredentialDefinition,
        preview: List[ProposedAttrib] = None, translation: List[AttribTranslation] = None, cred_id: str = None
):
    async with sirius_sdk.context(uri, credentials, p2p):
        machine = Issuer(holder=holder)
        success = await machine.issue(
            values=values,
            schema=schema,
            cred_def=cred_def,
            comment='Hello Iam issuer',
            preview=preview,
            translation=translation,
            cred_id=cred_id
        )
        return success


async def run_holder(
        uri: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        issuer: sirius_sdk.Pairwise, master_secret_id: str
):
    async with sirius_sdk.context(uri, credentials, p2p):
        listener = await sirius_sdk.subscribe()
        event = await listener.get_one()
        assert event.pairwise is not None
        offer = event.message
        assert isinstance(offer, OfferCredentialMessage)
        ttl = 60
        if offer.expires_time:
            expire = str_to_utc(offer.expires_time, raise_exceptions=False)
            delta = expire - datetime.utcnow()
            if delta.seconds > 0:
                ttl = delta.seconds
        machine = Holder(issuer=issuer, time_to_live=ttl)
        success, cred_id = await machine.accept(offer, master_secret_id, comment='Hello, Iam holder')
        return success, cred_id


async def run_issuer_indy_agent(
        indy_agent: IndyAgent, cred_def_id: str, cred_def: dict, values: dict, their_did: str,
        issuer_schema: dict = None, preview: List[ProposedAttrib] = None, translation: List[AttribTranslation] = None,
        rev_reg_id: str = None, cred_id: str = None, ttl: int = 60
) -> bool:
    log = await indy_agent.issue_credential(
        cred_def_id, cred_def, values, their_did, 'Test issuer', None, issuer_schema, preview, translation,
        rev_reg_id, cred_id, ttl
    )
    if len(log) > 2:
        last = log[-1]['message']
        pred = log[-2]['message']
        return ('Received ACK' in pred) and ('Done' in last)
    else:
        return False


@pytest.mark.asyncio
async def test_sane(
        test_suite: ServerTestSuite, agent1: sirius_sdk.Agent, agent2: sirius_sdk.Agent, prover_master_secret_name: str
):
    issuer = agent1
    holder = agent2
    await issuer.open()
    await holder.open()
    try:
        i2h = await get_pairwise(issuer, holder)
        h2i = await get_pairwise(holder, issuer)

        did_issuer, verkey_issuer = i2h.me.did, i2h.me.verkey
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True

        print('Prepare Holder')
        try:
            await holder.wallet.anoncreds.prover_create_master_secret(prover_master_secret_name)
        except AnoncredsMasterSecretDuplicateNameError:
            pass
    finally:
        await issuer.close()
        await holder.close()

    issuer = test_suite.get_agent_params('agent1')
    holder = test_suite.get_agent_params('agent2')
    holder_secret_id = prover_master_secret_name

    cred_id = 'cred-id-' + uuid.uuid4().hex
    coro_issuer = run_issuer(
        issuer['server_address'], issuer['credentials'], issuer['p2p'],
        holder=i2h,
        values={'attr1': 'Value-1', 'attr2': 567, 'attr3': 5.7},
        schema=schema, cred_def=cred_def, cred_id=cred_id
    )
    coro_holder = run_holder(
        holder['server_address'], holder['credentials'], holder['p2p'],
        issuer=h2i,
        master_secret_id=holder_secret_id
    )

    results = await run_coroutines(coro_issuer, coro_holder, timeout=60)
    cred_id = None
    for res in results:
        if type(res) is tuple:
            ok, cred_id = res
        else:
            ok = res
        assert ok is True

    assert cred_id is not None
    async with sirius_sdk.context(holder['server_address'], holder['credentials'], holder['p2p']):
        cred = await sirius_sdk.AnonCreds.prover_get_credential(cred_id)
    assert cred


@pytest.mark.asyncio
async def test_issuer_back_compatibility(indy_agent: IndyAgent, test_suite: ServerTestSuite, agent1: sirius_sdk.Agent):
    issuer = agent1
    await issuer.open()
    try:
        endpoint_issuer = [e for e in issuer.endpoints if e.routing_keys == []][0].address
        did_issuer, verkey_issuer = await issuer.wallet.did.create_and_store_my_did()
        did_holder, verkey_holder = await indy_agent.create_and_store_my_did()
        pairwise_for_issuer = sirius_sdk.Pairwise(
            me=sirius_sdk.Pairwise.Me(did_issuer, verkey_issuer),
            their=sirius_sdk.Pairwise.Their(did_holder, 'Holder', indy_agent.endpoint['url'], verkey_holder)
        )
        pairwise_for_holder = sirius_sdk.Pairwise(
            me=sirius_sdk.Pairwise.Me(did_holder, verkey_holder),
            their=sirius_sdk.Pairwise.Their(did_issuer, 'Issuer', endpoint_issuer, verkey_issuer)
        )
        pairwise_for_issuer.their.netloc = pytest.old_agent_overlay_address.replace('http://', '')
        pairwise_for_holder.their.netloc = pytest.test_suite_overlay_address.replace('http://', '')
        await indy_agent.create_pairwise_statically(pairwise_for_holder)
        await issuer.wallet.did.store_their_did(did_holder, verkey_holder)
        await issuer.pairwise_list.ensure_exists(pairwise_for_issuer)

        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, resp = await issuer.wallet.ledger.write_nym(
            'default', 'Th7MpTaRZVRYnPiabds81Y', did_issuer, verkey_issuer, 'Issuer', NYMRole.TRUST_ANCHOR
        )
        assert ok is True
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True
    finally:
        await issuer.close()

    issuer = test_suite.get_agent_params('agent1')
    cred_id = 'cred-id-' + uuid.uuid4().hex
    coro_issuer = run_issuer(
        issuer['server_address'], issuer['credentials'], issuer['p2p'],
        holder=pairwise_for_issuer,
        values={'attr1': 'Value-1', 'attr2': 567, 'attr3': 5.7},
        schema=schema, cred_def=cred_def, cred_id=cred_id
    )
    results = await run_coroutines(coro_issuer, timeout=60)
    assert len(results) == 1
    assert results[0] is True


@pytest.mark.asyncio
async def test_holder_back_compatibility(
        indy_agent: IndyAgent, agent1: sirius_sdk.Agent,
        test_suite: ServerTestSuite, prover_master_secret_name: str
):
    holder = agent1
    await holder.open()
    try:
        endpoint_holder = [e for e in holder.endpoints if e.routing_keys == []][0].address
        did_issuer, verkey_issuer = await indy_agent.create_and_store_my_did()
        did_holder, verkey_holder = await holder.wallet.did.create_and_store_my_did()
        pairwise_for_issuer = sirius_sdk.Pairwise(
            me=sirius_sdk.Pairwise.Me(did_issuer, verkey_issuer),
            their=sirius_sdk.Pairwise.Their(did_holder, 'Holder', endpoint_holder, verkey_holder)
        )
        pairwise_for_holder = sirius_sdk.Pairwise(
            me=sirius_sdk.Pairwise.Me(did_holder, verkey_holder),
            their=sirius_sdk.Pairwise.Their(did_issuer, 'Issuer', indy_agent.endpoint['url'], verkey_issuer)
        )
        pairwise_for_issuer.their.netloc = pytest.test_suite_overlay_address.replace('http://', '')
        pairwise_for_holder.their.netloc = pytest.old_agent_overlay_address.replace('http://', '')
        await indy_agent.create_pairwise_statically(pairwise_for_issuer)
        await holder.wallet.did.store_their_did(did_issuer, verkey_issuer)
        await holder.pairwise_list.ensure_exists(pairwise_for_holder)

        ok, resp = await agent1.wallet.ledger.write_nym(
            'default', 'Th7MpTaRZVRYnPiabds81Y', did_issuer, verkey_issuer, 'Issuer', NYMRole.TRUST_ANCHOR
        )
        assert ok is True
        # Register schema
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, schema = await indy_agent.register_schema(did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3'])
        cred_def_id, cred_def = await indy_agent.register_cred_def(did_issuer, schema_id, 'TAG')

        print('Prepare Holder')
        try:
            await holder.wallet.anoncreds.prover_create_master_secret(prover_master_secret_name)
        except AnoncredsMasterSecretDuplicateNameError:
            pass
    finally:
        await holder.close()

    holder = test_suite.get_agent_params('agent1')
    holder_secret_id = prover_master_secret_name
    cred_id = 'cred-id-' + uuid.uuid4().hex

    coro_issuer = run_issuer_indy_agent(
        indy_agent=indy_agent, cred_def_id=cred_def_id, cred_def=cred_def,
        values={'attr1': 'Value-1', 'attr2': 567, 'attr3': 5.7},
        their_did=pairwise_for_issuer.their.did,
        issuer_schema=schema, rev_reg_id=None,
        cred_id=cred_id
    )
    coro_holder = run_holder(
        holder['server_address'], holder['credentials'], holder['p2p'],
        issuer=pairwise_for_holder,
        master_secret_id=holder_secret_id
    )

    results = await run_coroutines(coro_issuer, coro_holder, timeout=60)
    assert len(results) == 2

    for res in results:
        if type(res) is tuple:
            ok, cred_id = res
        else:
            ok = res
        assert ok is True

    assert cred_id is not None
    async with sirius_sdk.context(holder['server_address'], holder['credentials'], holder['p2p']):
        cred = await sirius_sdk.AnonCreds.prover_get_credential(cred_id)
    assert cred



import datetime
from typing import Optional

from sirius_sdk.agent.aries_rfc.mixins import ThreadMixin, PleaseAckMixin
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage


class Message(ThreadMixin, PleaseAckMixin, AriesProtocolMessage, metaclass=RegisterMessage):
    """Implementation of BasicMessage protocol

    https://github.com/hyperledger/aries-rfcs/tree/master/features/0095-basic-message
    """

    PROTOCOL = 'basicmessage'
    NAME = 'message'

    def __init__(self, content: Optional[str] = None, locale: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if locale is not None:
            self['~l10n'] = {"locale": locale}
        if content is not None:
            self['content'] = content

    @property
    def content(self) -> Optional[str]:
        return self.get('content', None)

    @property
    def locale(self) -> Optional[str]:
        return self.get('~l10n', {}).get('locale', None)

    @property
    def sent_time(self) -> Optional[str]:
        return self.get('sent_time', None)

    def set_time(self):
        self['sent_time'] = datetime.datetime.now().astimezone().replace(microsecond=0).isoformat()



import json
from typing import Sequence
from collections import OrderedDict

from sirius_sdk.errors.exceptions import SiriusCryptoError
from sirius_sdk.encryption.custom import *


def ensure_is_bytes(b58_or_bytes: Union[str, bytes]) -> bytes:
    if isinstance(b58_or_bytes, str):
        return b58_to_bytes(b58_or_bytes)
    else:
        return b58_or_bytes


def prepare_pack_recipient_keys(
        to_verkeys: Sequence[bytes],
        from_verkey: bytes = None,
        from_sigkey: bytes = None
) -> (str, bytes):
    """
    Assemble the recipients block of a packed message.

    :param to_verkeys: Verkeys of recipients
    :param from_verkey: Sender Verkey needed to authcrypt package
    :param from_sigkey: Sender Sigkey needed to authcrypt package
    :return A tuple of (json result, key)
    """
    if from_verkey is not None and from_sigkey is None or \
            from_sigkey is not None and from_verkey is None:
        raise SiriusCryptoError(
            'Both verkey and sigkey needed to authenticated encrypt message'
        )

    cek = nacl.bindings.crypto_secretstream_xchacha20poly1305_keygen()
    recips = []

    for target_vk in to_verkeys:
        target_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(
            target_vk
        )
        if from_verkey:
            sender_vk = bytes_to_b58(from_verkey).encode("ascii")
            enc_sender = nacl.bindings.crypto_box_seal(sender_vk, target_pk)
            sk = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(
                from_sigkey
            )

            nonce = nacl.utils.random(nacl.bindings.crypto_box_NONCEBYTES)
            enc_cek = nacl.bindings.crypto_box(cek, nonce, target_pk, sk)
        else:
            enc_sender = None
            nonce = None
            enc_cek = nacl.bindings.crypto_box_seal(cek, target_pk)

        recips.append(
            OrderedDict(
                [
                    ("encrypted_key", bytes_to_b64(enc_cek, urlsafe=True)),
                    (
                        "header",
                        OrderedDict(
                            [
                                ("kid", bytes_to_b58(target_vk)),
                                (
                                    "sender",
                                    bytes_to_b64(enc_sender, urlsafe=True)
                                    if enc_sender
                                    else None,
                                ),
                                (
                                    "iv",
                                    bytes_to_b64(nonce, urlsafe=True)
                                    if nonce
                                    else None,
                                ),
                            ]
                        ),
                    ),
                ]
            )
        )

    data = OrderedDict(
        [
            ("enc", "xchacha20poly1305_ietf"),
            ("typ", "JWM/1.0"),
            ("alg", "Authcrypt" if from_verkey else "Anoncrypt"),
            ("recipients", recips),
        ]
    )
    return json.dumps(data), cek


def locate_pack_recipient_key(
        recipients: Sequence[dict],
        my_verkey: bytes,
        my_sigkey: bytes
) -> (bytes, str, str):
    """
    Locate pack recipient key.

    Decode the encryption key and sender verification key from a
    corresponding recipient block, if any is defined.

    :param recipients: Recipients to locate
    :param my_verkey: Verkey needed to auth-decrypt
    :param my_sigkey: Sigkey needed to auth-decrypt
    :return A tuple of (cek, sender_vk, recip_vk_b58)

    Raises: ValueError: If no corresponding recipient key found
    """
    not_found = []
    for recip in recipients:
        if not recip or "header" not in recip or "encrypted_key" not in recip:
            raise ValueError("Invalid recipient header")

        recip_vk_b58 = recip["header"].get("kid")

        if bytes_to_b58(my_verkey) != recip_vk_b58:
            not_found.append(recip_vk_b58)
            continue

        pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(my_verkey)
        sk = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(my_sigkey)

        encrypted_key = b64_to_bytes(recip["encrypted_key"], urlsafe=True)

        if "iv" in recip["header"] and recip["header"]["iv"] and \
                "sender" in recip["header"] and recip["header"]["sender"]:
            nonce = b64_to_bytes(recip["header"]["iv"], urlsafe=True)
            enc_sender = b64_to_bytes(recip["header"]["sender"], urlsafe=True)
        else:
            nonce = None
            enc_sender = None

        if nonce and enc_sender:
            sender_vk = nacl.bindings.crypto_box_seal_open(
                enc_sender,
                pk,
                sk
            ).decode("ascii")
            sender_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(
                b58_to_bytes(sender_vk)
            )
            cek = nacl.bindings.crypto_box_open(
                encrypted_key,
                nonce,
                sender_pk,
                sk
            )
        else:
            sender_vk = None
            cek = nacl.bindings.crypto_box_seal_open(encrypted_key, pk, sk)
        return cek, sender_vk, recip_vk_b58
    raise ValueError(
        "No corresponding recipient key found in {}".format(not_found)
    )


def encrypt_plaintext(
        message: str, add_data: bytes, key: bytes
) -> (bytes, bytes, bytes):
    """
    Encrypt the payload of a packed message.

    :param message: Message to encrypt
    :param add_data: additional data
    :param key: Key used for encryption
    :return A tuple of (ciphertext, nonce, tag)
    """
    nonce = nacl.utils.random(
        nacl.bindings.crypto_aead_chacha20poly1305_ietf_NPUBBYTES
    )
    message_bin = message.encode("ascii")
    output = nacl.bindings.crypto_aead_chacha20poly1305_ietf_encrypt(
        message_bin, add_data, nonce, key
    )
    mlen = len(message)
    ciphertext = output[:mlen]
    tag = output[mlen:]
    return ciphertext, nonce, tag


def decrypt_plaintext(
        ciphertext: bytes, recips_bin: bytes, nonce: bytes, key: bytes
) -> str:
    """
    Decrypt the payload of a packed message.

    :param ciphertext
    :param recips_bin
    :param nonce
    :param key
    :return The decrypted string
    """
    output = nacl.bindings.crypto_aead_chacha20poly1305_ietf_decrypt(
        ciphertext, recips_bin, nonce, key
    )
    return output.decode("ascii")


def pack_message(
        message: str,
        to_verkeys: Sequence[Union[bytes, str]],
        from_verkey: Union[bytes, str] = None,
        from_sigkey: Union[bytes, str] = None
) -> bytes:
    """
    Assemble a packed message for a set of recipients, optionally including
    the sender.

    :param message: The message to pack
    :param to_verkeys: (Sequence of bytes or base58 string) The verkeys to pack the message for
    :param from_verkey: (bytes or base58 string) The sender verkey
    :param from_sigkey: (bytes or base58 string) The sender sigkey
    :return The encoded message
    """

    to_verkeys = [ensure_is_bytes(vk) for vk in to_verkeys]
    from_verkey = ensure_is_bytes(from_verkey)
    from_sigkey = ensure_is_bytes(from_sigkey)

    recips_json, cek = prepare_pack_recipient_keys(
        to_verkeys,
        from_verkey,
        from_sigkey
    )
    recips_b64 = bytes_to_b64(recips_json.encode("ascii"), urlsafe=True)

    ciphertext, nonce, tag = encrypt_plaintext(
        message,
        recips_b64.encode("ascii"),
        cek
    )

    data = OrderedDict(
        [
            ("protected", recips_b64),
            ("iv", bytes_to_b64(nonce, urlsafe=True)),
            ("ciphertext", bytes_to_b64(ciphertext, urlsafe=True)),
            ("tag", bytes_to_b64(tag, urlsafe=True)),
        ]
    )
    return json.dumps(data).encode("ascii")


def unpack_message(
        enc_message: Union[bytes, dict], my_verkey: Union[bytes, str], my_sigkey: Union[bytes, str]
) -> (str, Optional[str], str):
    """
    Decode a packed message.

    Disassemble and unencrypt a packed message, returning the message content,
    verification key of the sender (if available), and verification key of the
    recipient.

    :param enc_message: The encrypted message
    :param my_verkey: (bytes or base58 string) Verkey for decrypt
    :param my_sigkey: (bytes or base58 string) Sigkey for decrypt
    :return A tuple of (message, sender_vk, recip_vk)
    Raises:
        ValueError: If the packed message is invalid
        ValueError: If the packed message reipients are invalid
        ValueError: If the pack algorithm is unsupported
        ValueError: If the sender's public key was not provided

    """
    my_verkey = ensure_is_bytes(my_verkey)
    my_sigkey = ensure_is_bytes(my_sigkey)

    if not isinstance(enc_message, bytes) and \
            not isinstance(enc_message, dict):
        raise TypeError(
            'Expected bytes or dict, got {}'.format(type(enc_message))
        )
    if isinstance(enc_message, bytes):
        try:
            enc_message = json.loads(enc_message)
        except Exception as err:
            raise ValueError("Invalid packed message") from err

    protected_bin = enc_message["protected"].encode("ascii")
    recips_json = b64_to_bytes(
        enc_message["protected"], urlsafe=True
    ).decode("ascii")
    try:
        recips_outer = json.loads(recips_json)
    except Exception as err:
        raise ValueError("Invalid packed message recipients") from err

    alg = recips_outer["alg"]
    is_authcrypt = alg == "Authcrypt"
    if not is_authcrypt and alg != "Anoncrypt":
        raise ValueError("Unsupported pack algorithm: {}".format(alg))
    cek, sender_vk, recip_vk = locate_pack_recipient_key(
        recips_outer["recipients"], my_verkey, my_sigkey
    )
    if not sender_vk and is_authcrypt:
        raise ValueError(
            "Sender public key not provided for Authcrypt message"
        )

    ciphertext = b64_to_bytes(enc_message["ciphertext"], urlsafe=True)
    nonce = b64_to_bytes(enc_message["iv"], urlsafe=True)
    tag = b64_to_bytes(enc_message["tag"], urlsafe=True)

    payload_bin = ciphertext + tag
    message = decrypt_plaintext(payload_bin, protected_bin, nonce, cek)

    return message, sender_vk, recip_vk
